"""CLI init function."""

# These imports are necessary to register CLI commands. Do not remove!
from .benchmark_cli import download_data  # noqa: F401
from .cli import cli  # noqa: F401
from .fine_tune_cli import finetune  # noqa: F401
from .segment_cli import segment  # noqa: F401
from .ske_cli import skeletonize  # noqa: F401
from .train_cli import data_dir_help, train  # noqa: F401



"""membrane segmentation in 3D for cryo-ET."""

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("membrain-seg")
except PackageNotFoundError:
    __version__ = "uninstalled"

__author__ = "Lorenz Lamm"
__email__ = "lorenz.lamm@helmholtz-munich.de"



from typing import Any, Callable, List, Optional, Tuple, Union

import numpy as np
import torch
from monai.transforms import (
    Compose,
    MapTransform,
    OneOf,
    RandAdjustContrastd,
    Randomizable,
    Resize,
    Transform,
    Transposed,
    Zoomd,
)
from scipy.ndimage import convolve, median_filter


class RandApplyTransform(Randomizable, Transform):
    """Randomly apply given MONAI transform with given probability.

    This class is a wrapper around a MONAI transform. It extends the
    Randomizable and Transform classes provided by MONAI. The purpose of
    this class is to randomly apply a specific MONAI transform to the data
    based on a given probability. For example, if the given probability
    is 0.5, then the wrapped transform will be applied to the data 50%
    of the time when called.
    """

    def __init__(
        self,
        transform: Transform,
        prob: float = 0.5,
        do_transform: Optional[bool] = None,
    ):
        if prob < 0.0 or prob > 1.0:
            raise ValueError("Probability must be a value between 0 and 1.")
        self.transform = transform
        self.prob = prob
        self.do_transform = do_transform

    def randomize(self, data: None) -> None:
        """Roll a dice and decide whether we actuall do the augmentation."""
        if self.do_transform is None:
            self._do_transform = self.R.random() < self.prob
        else:
            self._do_transform = self.do_transform

    def __call__(self, data):
        """Apply transform only with given probability."""
        self.randomize(None)
        if self._do_transform:
            return self.transform(data)
        return data


class MedianFilterd(Transform):
    """Median filter from batchgenerators package reimplemented.

    This class extends the Transform class provided by MONAI and implements a median
    filter operation for image data. A median filter is a digital filtering technique,
    used to remove noise from an image or signal. The class allows the radius of
    the median filter to be either a specific integer or a range from which an integer
    can be randomly selected each time the transform is called.
    """

    def __init__(self, keys, radius=1):
        self.keys = keys
        self.radius_range = radius

    def randomize(self, data: None) -> None:
        """Randomly draw median filter range."""
        if isinstance(self.radius_range, int):
            self.radius = self.radius_range
        elif isinstance(self.radius_range, tuple):
            self.radius = np.random.randint(self.radius_range[0], self.radius_range[1])

    def __call__(self, data):
        """Apply median filter."""
        self.randomize(None)
        for key in self.keys:
            for c in range(data[key].shape[0]):
                data[key][c] = torch.from_numpy(
                    median_filter(data[key][c], size=self.radius, mode="reflect")
                )  # TODO: Is this very inefficient?
        return data


class RandomBrightnessTransformd(Transform):
    """Brightness transform from batchgenerators reimplemented.

    This class extends the Transform class provided by MONAI and is used to
    apply a random brightness shift to the image data. The extent of the
    brightness shift follows a Gaussian (normal) distribution defined by the
    given mean (mu) and standard deviation (sigma). The transform is applied based
    on a given probability. For example, if the given probability is 0.7, the brightness
    shift will be applied to the image 70% of the time when the transform is called.
    """

    def __init__(self, keys, mu, sigma, prob=1.0):
        super().__init__()
        self.keys = keys
        self.mu = mu
        self.sigma = sigma
        self.prob = prob

    def __call__(self, data):
        """Apply brightness transform."""
        if np.random.rand() < self.prob:
            for key in self.keys:
                add_const = np.random.normal(loc=self.mu, scale=self.sigma)
                data[key] = data[key] + add_const
        return data


class RandomContrastTransformd(Transform):
    """Randomly scale the data with a contrast factor.

    This class extends the Transform class provided by MONAI. It's designed to randomly
    scale the contrast of an image. It allows to specify the keys to the images to be
    transformed, a range for contrast scaling, and whether the original range of the
    data should be preserved after the transformation.
    The probability of applying this transformation can also be set.
    """

    def __init__(self, keys, contrast_range, preserve_range=False, prob=1.0):
        super().__init__()
        self.keys = keys
        self.contrast_range = contrast_range
        self.prob = prob
        self.preserve_range = preserve_range

    def __call__(self, data):
        """Apply the transform."""
        if np.random.rand() < self.prob:
            for key in self.keys:
                # Choose a random contrast factor from the given range
                contrast_factor = np.random.uniform(
                    self.contrast_range[0], self.contrast_range[1]
                )

                # Apply the contrast transformation
                mean = np.mean(data[key])
                if self.preserve_range:
                    minval = data[key].min()
                    maxval = data[key].max()
                data[key] = mean + contrast_factor * (data[key] - mean)
                if self.preserve_range:
                    data[key][data[key] < minval] = minval
                    data[key][data[key] > maxval] = maxval

        return data


class RandAdjustContrastWithInversionAndStats(RandAdjustContrastd):
    """
    Adjusted version of RandAdjustContrastd from MONAI.

    After the normal function call, the original mean and std are preserved
    and the result is multiplied by -1. It is therefore recommended to
    apply the function 2 times.
    (Without two time application and negation, the transform is non-symmetric)
    """

    def __call__(self, data):
        """Apply the transform."""
        for key in self.keys:
            if self._do_transform:
                img = data[key]
                mean_before = img.mean()
                std_before = img.std()

                img = super().__call__({key: img})[key]

                mean_after = img.mean()
                std_after = img.std()

                img = (img - mean_after) / std_after * std_before + mean_before
                img *= -1
                data[key] = img

        return data


class SimulateLowResolutionTransform(Randomizable, Transform):
    """
    Batchgenerators' transform for simulating low resolution upscaling.

    The image is downsampled using nearest neighbor interpolation using a
    random scale factor. Afterwards, it is upsampled again with trilinear
    interpolation, imitating the upscaling of low-resolution images.
    """

    def __init__(
        self,
        keys,
        downscale_factor_range,
        upscale_mode="trilinear",
        downscale_mode="nearest",
    ):
        self.keys = keys
        self.downscale_factor_range = downscale_factor_range
        self.downscale_factor = None
        self.upscale_mode = upscale_mode
        self.downscale_mode = downscale_mode

    def randomize(self, data):
        """Randomly choose the downsample factor."""
        self.downscale_factor = self.R.uniform(*self.downscale_factor_range)

    def __call__(self, data):
        """Apply the transform."""
        self.randomize(data)
        downscale_transform = Zoomd(
            self.keys, zoom=self.downscale_factor, mode=self.downscale_mode
        )
        upscale_transform = Zoomd(
            self.keys, zoom=1.0 / self.downscale_factor, mode=self.upscale_mode
        )

        data = downscale_transform(data)
        data = upscale_transform(data)
        return data


class BlankCuboidTransform(Randomizable, MapTransform):
    """
    Randomly blank out cuboids from the image.

    It's designed to randomly blank out one or more cuboid regions in an image.
    The number and size of the cuboids are random, with the sizes drawn from a specified
    range. The class also provides options to replace the blanked regions with
    either the mean value of the image or zeros.
    """

    def __init__(
        self,
        keys,
        prob=0.5,
        cuboid_area=(10, 100),
        is_3d=False,
        max_cuboids=1,
        replace_with="mean",
    ):
        super().__init__(keys)
        self.prob = prob
        self.cuboid_area = cuboid_area
        self.is_3d = is_3d
        self.max_cuboids = max_cuboids
        self.replace_with = replace_with

    def randomize(self, data=None):
        """Randomly choose the number of cuboids & their sizes."""
        self.do_transform = self.R.random() < self.prob
        if self.do_transform:
            self.num_cuboids = self.R.randint(1, self.max_cuboids + 1)
            self.cuboids = []
            for _ in range(self.num_cuboids):
                height = self.R.randint(self.cuboid_area[0], self.cuboid_area[1] + 1)
                width = self.R.randint(self.cuboid_area[0], self.cuboid_area[1] + 1)
                depth = (
                    self.R.randint(self.cuboid_area[0], self.cuboid_area[1] + 1)
                    if self.is_3d
                    else 1
                )
                self.cuboids.append((height, width, depth))

    def __call__(self, data):
        """Apply the transform."""
        self.randomize()
        if not self.do_transform:
            return data

        d = dict(data)
        for key in self.keys:
            image = d[key]
            for height, width, depth in self.cuboids:
                if self.is_3d:
                    z_max, y_max, x_max = (
                        image.shape[-3],
                        image.shape[-2],
                        image.shape[-1],
                    )
                    z = self.R.randint(0, z_max - depth)
                else:
                    y_max, x_max = image.shape[-2], image.shape[-1]
                    z, depth = 0, 1
                y = self.R.randint(0, y_max - height)
                x = self.R.randint(0, x_max - width)
                if self.replace_with == "mean":
                    image[..., z : z + depth, y : y + height, x : x + width] = (
                        torch.mean(torch.Tensor(image))
                    )
                elif self.replace_with == 0.0:
                    image[..., z : z + depth, y : y + height, x : x + width] = 0.0
            d[key] = image

        return d


def sample_scalar(value: Union[Tuple, List, Callable, Any], *args: Any) -> Any:
    """
    Implementation from the batchgenerators package.

    Function to sample a scalar from a specified range, or compute it using
    a provided function.

    Args:
        value: The value to be sampled. It can be a tuple/list defining a range,
               a callable function, or any other value. If it's a tuple/list,
               a random value within the range is returned. If it's a function, it is
               called with the arguments supplied in *args. If it's any other value,
               it is returned as is.
        *args: Additional arguments to be passed to the callable 'value', if it is a
                function.

    Returns
    -------
        A sampled or computed scalar value.
    """
    if isinstance(value, (tuple, list)):
        return np.random.uniform(value[0], value[1])
    elif callable(value):
        # return value(image, kernel)
        return value(*args)
    else:
        return value


class BrightnessGradientAdditiveTransform(Randomizable, MapTransform):
    """
    Add a brightness gradient to the image (also from batchgenerators).

    A scaled Gaussian kernel is added to the image. The center of the
    Gaussian kernel is randomly drawn an can also be outside of the image.
    """

    def __init__(
        self,
        keys,
        scale,
        loc=(-1, 2),
        max_strength=1.0,
        mean_centered=True,
    ):
        super().__init__(keys)
        self.scale = scale
        self.loc = loc
        self.max_strength = max_strength
        self.mean_centered = mean_centered

    def _generate_kernel(self, img_shape):
        img_shape = img_shape[2:]
        scale = [sample_scalar(self.scale, img_shape, i) for i in range(len(img_shape))]
        loc = [sample_scalar(self.loc, img_shape, i) for i in range(len(img_shape))]
        coords = [
            np.linspace(-loc[i], img_shape[i] - loc[i], img_shape[i])
            for i in range(len(img_shape))
        ]
        meshgrid = np.meshgrid(*coords, indexing="ij")
        kernel = np.exp(
            -0.5 * sum([(meshgrid[i] / scale[i]) ** 2 for i in range(len(img_shape))])
        )
        kernel = np.expand_dims(kernel, 0)
        return kernel

    def __call__(self, data):
        """Apply the transform."""
        d = dict(data)
        for key in self.keys:
            image = d[key]
            img_shape = image.shape
            kernel = self._generate_kernel(img_shape)
            if self.mean_centered:
                kernel -= kernel.mean()
            max_kernel_val = max(np.max(np.abs(kernel)), 1e-8)
            strength = sample_scalar(self.max_strength, image, kernel)
            kernel = kernel / max_kernel_val * strength
            image += kernel

            d[key] = image

        return d


class LocalGammaTransform(Randomizable, MapTransform):
    """Locally adjusts the Gamma value using Gaussian kernel.(from batchgenerators).

    This transform adjusts the Gamma value of the images locally. The Gamma
    correction is a type of power-law transformation that manipulates the
    brightness of an image. It's applied locally based on a generated Gaussian
    kernel. This class is an implementation from the batchgenerators package.

    """

    def __init__(self, keys, scale, loc=(-1, 2), gamma=(0.5, 1)):
        super().__init__(keys)
        self.scale = scale
        self.loc = loc
        self.gamma = gamma

    def _generate_kernel(self, img_shape):
        img_shape = img_shape[2:]
        scale = [sample_scalar(self.scale, img_shape, i) for i in range(len(img_shape))]
        loc = [sample_scalar(self.loc, img_shape, i) for i in range(len(img_shape))]
        coords = [
            np.linspace(-loc[i], img_shape[i] - loc[i], img_shape[i])
            for i in range(len(img_shape))
        ]
        meshgrid = np.meshgrid(*coords, indexing="ij")
        kernel = np.exp(
            -0.5 * sum([(meshgrid[i] / scale[i]) ** 2 for i in range(len(img_shape))])
        )
        kernel = np.expand_dims(kernel, 0)
        return kernel

    def _apply_gamma_gradient(self, img, kernel):
        mn, mx = img.min(), img.max()
        img = (img - mn) / (max(mx - mn, 1e-8))

        gamma = sample_scalar(self.gamma)
        img_modified = np.power(img, gamma)

        return self.run_interpolation(img, img_modified, kernel) * (mx - mn) + mn

    def run_interpolation(self, img, img_modified, kernel):
        """Interpolate between image and modified image, weighted with kernel."""
        return img + kernel * (img_modified - img)

    def __call__(self, data):
        """Apply the transform."""
        d = dict(data)
        for key in self.keys:
            image = d[key]
            img_shape = image.shape
            kernel = self._generate_kernel(img_shape)
            d[key] = self._apply_gamma_gradient(image, kernel)
        return d


class SharpeningTransformMONAI(MapTransform):
    """Laplacian Sharpening transform. (from batchgenerators).

    This transform applies a Laplacian sharpening filter to the image, enhancing
    details and emphasizing edges. The strength of sharpening can be controlled
    and can be set to be the same or different for each image channel. This class
    is an implementation from the batchgenerators package.

    """

    filter_2d = np.array([[0, -1, 0], [-1, 4, -1], [0, -1, 0]])
    filter_3d = np.array(
        [
            [[0, 0, 0], [0, -1, 0], [0, 0, 0]],
            [[0, -1, 0], [-1, 6, -1], [0, -1, 0]],
            [[0, 0, 0], [0, -1, 0], [0, 0, 0]],
        ]
    )

    def __init__(
        self,
        strength: Union[float, Tuple[float, float]] = 0.2,
        same_for_each_channel: bool = False,
        p_per_channel: float = 0.5,
        keys: str = "image",
    ):
        super().__init__(keys)
        self.strength = strength
        self.same_for_each_channel = same_for_each_channel
        self.p_per_channel = p_per_channel

    def __call__(self, data):
        """Apply the transform."""
        d = dict(data)
        for key in self.keys:
            img = d[key]
            if len(img.shape) == 2:
                img = np.expand_dims(img, axis=0)
            if self.same_for_each_channel:
                strength_here = (
                    self.strength
                    if isinstance(self.strength, float)
                    else np.random.uniform(*self.strength)
                )
                filter_here = (
                    self.filter_2d if img.ndim == 3 else self.filter_3d
                ) * strength_here
                filter_here[tuple([1] * (img.ndim - 1))] += 1
                for c in range(img.shape[0]):
                    if np.random.uniform() < self.p_per_channel:
                        img[c] = torch.from_numpy(
                            convolve(img[c], filter_here, mode="reflect")
                        )
                        img[c] = np.clip(img[c], img[c].min(), img[c].max())
            else:
                for c in range(img.shape[0]):
                    if np.random.uniform() < self.p_per_channel:
                        strength_here = (
                            self.strength
                            if isinstance(self.strength, float)
                            else np.random.uniform(*self.strength)
                        )
                        filter_here = (
                            self.filter_2d if img.ndim == 3 else self.filter_3d
                        ) * strength_here
                        filter_here[tuple([1] * (img.ndim - 1))] += 1
                        img[c] = torch.from_numpy(
                            convolve(img[c], filter_here, mode="reflect")
                        )
                        img[c] = torch.from_numpy(
                            np.clip(img[c], img[c].min(), img[c].max())
                        )
            d[key] = img.squeeze() if len(img.shape) == 5 and img.shape[0] == 1 else img
        return d


class DownsampleSegForDeepSupervisionTransform(MapTransform):
    """Downsample labels for deep supervision (from nnUNet).

    This transform downsamples the given labels for the purpose of deep supervision
    during model training. Deep supervision involves making predictions at
    multiple scales or layers, which can help improve performance. This class
    is an implementation from the nnUNet package.
    """

    def __init__(
        self,
        keys: Union[str, Tuple[str]],
        ds_scales: Tuple[float, float, float] = (1, 0.5, 0.25),
        order: int = 0,
        axes: Union[None, Tuple[int, int, int]] = None,
    ):
        super().__init__(keys)
        self.axes = axes
        self.order = order
        self.ds_scales = ds_scales

    def __call__(self, data_dict):
        """Apply the transform."""
        for key in self.keys:
            data_dict[key] = downsample_seg_for_ds_transform(
                data_dict[key], self.ds_scales, self.order, self.axes
            )
        return data_dict


def downsample_seg_for_ds_transform(
    seg: np.ndarray,
    ds_scales: Tuple[float, float, float] = (1, 0.5, 0.25),
    order: str = "nearest",
    axes: Union[None, Tuple[int, int, int]] = None,
):
    """
    Function for downsampling segmentations based on provided downscaling factors.

    For each downscaling factor in 'ds_scales', the input segmentation ('seg') is
    resized along the specified 'axes'. If 'axes' is None, then all axes excluding
    the first one (typically the channel axis) are resized. The interpolation mode
    for resizing is specified by 'order', which defaults to 'nearest'.

    Parameters
    ----------
    seg : np.ndarray
        Input segmentation to be downsampled.
    ds_scales : Tuple[float, float, float], optional
        Downscaling factors for each axis, by default (1, 0.5, 0.25).
    order : str, optional
        Interpolation mode to be used while resizing, by default 'nearest'.
    axes : Union[None, Tuple[int, int, int]], optional
        Axes along which the input is resized. If None, all axes excluding the
        first one are resized, by default None.

    Returns
    -------
    list
        A list of downsampled segmentations corresponding to each scale in 'ds_scales'.
    """
    if axes is None:
        axes = list(range(1, len(seg.shape)))
    output = []
    for s in ds_scales:
        if all([i == 1 for i in s]):
            output.append(seg)
        else:
            new_shape = np.array(seg.shape).astype(float)
            for i, a in enumerate(axes):
                new_shape[a] *= s[i]
            new_shape = np.round(new_shape).astype(int)
            resize_transform = Resize(
                new_shape[1:],
                mode=order,
                align_corners=(None if order == "nearest" else False),
            )
            out_seg = np.zeros(new_shape, dtype=np.array(seg).dtype)
            for c in range(seg.shape[0]):
                out_seg[c] = resize_transform(np.expand_dims(seg[c], 0)).squeeze()
            output.append(out_seg)
    return output


AxesSwap = OneOf(
    [
        Transposed(keys=("image", "label"), indices=(0, 2, 1, 3)),
        Transposed(keys=("image", "label"), indices=(0, 1, 3, 2)),
        Transposed(keys=("image", "label"), indices=(0, 3, 2, 1)),
    ]
)
AxesShuffle = Compose(
    [
        RandApplyTransform(transform=AxesSwap, prob=0.5),
        RandApplyTransform(transform=AxesSwap, prob=0.5),
        RandApplyTransform(transform=AxesSwap, prob=0.5),
    ]
)



# Derived from: Python utilities for Focus
# Author: Ricardo Righetto
# E-mail: ricardo.righetto@unibas.ch
# https://github.com/C-CINA/focustools/

import warnings

import numpy as np

warnings.filterwarnings("ignore", category=RuntimeWarning)

pi = np.pi  # global PI

# TO-DO:
# Port heavy calculations to Torch or something more efficient than pure NumPy?
# Original implementation used numexpr (see commented code) but that would add one more
# dependency to MemBrain and does not give significant speedups, at least not with the
# defaults.
# For now, we stick to pure Numpy.


def RadialIndices(
    imsize: tuple = (128, 128),
    rounding: bool = True,
    normalize: bool = False,
    rfft: bool = False,
    xyz: tuple = (0, 0, 0),
    nozero: bool = True,
    nozeroval: float = 1e-3,
):
    """
    Generates a 1D/2D/3D array whose values are the distance counted from the origin. 

    Parameters
    ----------
    imsize : tuple
        The shape of the input ndarray.
    rounding : bool
        Whether the radius values should be rounded to the nearest integer ensuring \
"perfect" radial symmetry.
    normalize : bool
        Whether the radius values should be normalized to the range [0.0,11.0].
    rfft : bool
        Whether to return an array consistent with np.fft.rfftn i.e. exploiting the \
Hermitian symmetry of the Fourier transform of real data.
    xyz : tuple
        Shifts to be applied to the origin specified as (x_shift, y_shift, z_shift). \
Useful when applying phase shifts.
    nozero : bool
        Whether the value of the origin (corresponding to the zero frequency or DC \
component in the Fourier transform) should be set to a small value instead of zero.
    nozeroval : float
        The value to put at the origin if nozero is True.

    Returns
    -------
    rmesh : ndarray
        Array whose values are the distance from the origin.
    amesh : ndarray
        Array whose values are the angle from the x- axis (2D) or from the x,y plane \
(3D)

    Raises
    ------
    ValueError
        If imsize with more than 3 dimensions is given.

    Notes
    -----
    This function is compliant with NumPy fft.fftfreq() and fft.rfftfreq().
    """
    imsize = np.array(imsize)

    # if np.isscalar(imsize):
    #     imsize = [imsize, imsize]

    if len(imsize) > 3:
        raise ValueError(
            "Object should have 2 or 3 dimensions: len(imsize) = %d " % len(imsize)
        )

    xyz = np.flipud(xyz)

    m = np.mod(imsize, 2)  # Check if dimensions are odd or even

    if len(imsize) == 1:
        # The definition below is consistent with numpy np.fft.fftfreq and
        # np.fft.rfftfreq:

        if not rfft:
            xmesh = np.mgrid[
                -imsize[0] // 2 + m[0] - xyz[0] : (imsize[0] - 1) // 2 + 1 - xyz[0]
            ]

        else:
            xmesh = np.mgrid[0 - xyz[0] : imsize[0] // 2 + 1 - xyz[0]]

        rmesh = np.sqrt(xmesh * xmesh)

        amesh = np.zeros(xmesh.shape)

        n = 1  # Normalization factor

    if len(imsize) == 2:
        # The definition below is consistent with numpy np.fft.fftfreq and
        # np.fft.rfftfreq:

        if not rfft:
            [xmesh, ymesh] = np.mgrid[
                -imsize[0] // 2 + m[0] - xyz[0] : (imsize[0] - 1) // 2 + 1 - xyz[0],
                -imsize[1] // 2 + m[1] - xyz[1] : (imsize[1] - 1) // 2 + 1 - xyz[1],
            ]

        else:
            [xmesh, ymesh] = np.mgrid[
                -imsize[0] // 2 + m[0] - xyz[0] : (imsize[0] - 1) // 2 + 1 - xyz[0],
                0 - xyz[1] : imsize[1] // 2 + 1 - xyz[1],
            ]
            xmesh = np.fft.ifftshift(xmesh)

        rmesh = np.sqrt(xmesh * xmesh + ymesh * ymesh)

        amesh = np.arctan2(ymesh, xmesh)

        n = 2  # Normalization factor

    if len(imsize) == 3:
        # The definition below is consistent with numpy np.fft.fftfreq and
        # np.fft.rfftfreq:

        if not rfft:
            [xmesh, ymesh, zmesh] = np.mgrid[
                -imsize[0] // 2 + m[0] - xyz[0] : (imsize[0] - 1) // 2 + 1 - xyz[0],
                -imsize[1] // 2 + m[1] - xyz[1] : (imsize[1] - 1) // 2 + 1 - xyz[1],
                -imsize[2] // 2 + m[2] - xyz[2] : (imsize[2] - 1) // 2 + 1 - xyz[2],
            ]

        else:
            [xmesh, ymesh, zmesh] = np.mgrid[
                -imsize[0] // 2 + m[0] - xyz[0] : (imsize[0] - 1) // 2 + 1 - xyz[0],
                -imsize[1] // 2 + m[1] - xyz[1] : (imsize[1] - 1) // 2 + 1 - xyz[1],
                0 - xyz[2] : imsize[2] // 2 + 1 - xyz[2],
            ]
            xmesh = np.fft.ifftshift(xmesh)
            ymesh = np.fft.ifftshift(ymesh)

        rmesh = np.sqrt(xmesh * xmesh + ymesh * ymesh + zmesh * zmesh)

        amesh = np.arccos(zmesh / rmesh)

        n = 3  # Normalization factor

    if rounding:
        rmesh = np.round(rmesh)

    if normalize:
        a = np.sum(imsize * imsize)

        rmesh = rmesh / (np.sqrt(a) / np.sqrt(n))

    if nozero:
        # Replaces the "zero radius" by a small value to prevent division by zero in
        # other programs
        idx = rmesh == 0
        rmesh[idx] = nozeroval

    return rmesh, np.nan_to_num(amesh)


def CTF(
    imsize: tuple = (128, 128),
    df1: float = 50000.0,
    df2: float = None,
    ast: float = 0.0,
    ampcon: float = 0.07,
    Cs: float = 2.7,
    kV: float = 300.0,
    apix: float = 1.0,
    B: float = 0.0,
    rfft: bool = True,
):
    """
    Generates 1D, 2D or 3D contrast transfer function (CTF) of a TEM.

    Parameters
    ----------
    imsize : tuple
        The shape of the input ndarray.
    df1 : float
        Defocus 1 (or Defocus U in some notations) in Angstroms. Principal defocus \
axis. Underfocus is positive.
    df2 : float
        Defocus 2 (or Defocus V in some notations) in Angstroms. Defocus axis \
orthogonal to the U axis. Only mandatory for astigmatic data.
    ast : float
        Angle for astigmatic data (in degrees).
    ampcon : float
        Amplitude contrast fraction (between 0.0 and 1.0).
    Cs : float
        Spherical aberration (in mm).
    kV : float
        Acceleration voltage of the TEM (in kV).
    apix : float
        Input pixel size in Angstroms.
    B : float
        B-factor in Angstroms**2.
    rfft : bool
        Whether to return an array consistent with np.fft.rfftn i.e. exploiting the \
Hermitian symmetry of the Fourier transform of real data.

    Returns
    -------
    CTFim : ndarray
        Array containing the CTF.

    Notes
    -----
    Follows the CTF definition from Mindell & Grigorieff, JSB (2003) \
(https://doi.org/10.1016/S1047-8477(03)00069-8), which is adopted in FREALIGN/\
cisTEM, RELION and many other packages.
    """
    if not np.isscalar(imsize) and len(imsize) == 1:
        imsize = imsize[0]

    Cs *= 1e7  # Convert Cs to Angstroms

    if df2 is None or np.isscalar(imsize):
        df2 = df1

    # NOTATION FOR DEFOCUS1, DEFOCUS2, ASTIGMATISM BELOW IS INVERTED DUE TO NUMPY
    # CONVENTION:
    # df1, df2 = df2, df1

    ast *= -pi / 180.0

    WL = ElectronWavelength(kV)

    w1 = np.sqrt(1 - ampcon * ampcon)
    w2 = ampcon

    import warnings

    with warnings.catch_warnings():
        warnings.filterwarnings("ignore", category=RuntimeWarning)

        if np.isscalar(imsize):
            if rfft:
                rmesh = np.fft.rfftfreq(imsize)
            else:
                rmesh = np.fft.fftfreq(imsize)
            amesh = 0.0

        else:
            rmesh, amesh = RadialIndices(imsize, normalize=True, rfft=rfft)

            rmesh2 = rmesh**2 / apix**2

        # From Mindell & Grigorieff, JSB 2003:
        df = 0.5 * (df1 + df2 + (df1 - df2) * np.cos(2.0 * (amesh - ast)))

        Xr = np.nan_to_num(pi * WL * rmesh2 * (df - 0.5 * WL * WL * rmesh2 * Cs))

    CTFim = -w1 * np.sin(Xr) - w2 * np.cos(Xr)

    if B != 0.0:  # Apply B-factor only if necessary:
        CTFim = CTFim * np.exp(-B * (rmesh2) / 4)

    return CTFim


def CorrectCTF(
    img=None,
    df1: float = 50000.0,
    df2: float = None,
    ast: float = 0.0,
    ampcon: float = 0.07,
    invert_contrast: bool = False,
    Cs: float = 2.7,
    kV: float = 300.0,
    apix: float = 1.0,
    phase_flip: bool = False,
    ctf_multiply: bool = False,
    wiener_filter: bool = False,
    C: float = 1.0,
    return_ctf: bool = False,
):
    """
    Applies different types of CTF correction to a 2D or 3D image.

    Parameters
    ----------
    img : 
        The input image to be corrected.
    df1 : float
        Defocus 1 (or Defocus U in some notations) in Angstroms. Principal defocus \
axis. Underfocus is positive.
    df2 : float
        Defocus 2 (or Defocus V in some notations) in Angstroms. Defocus axis \
orthogonal to the U axis. Only mandatory for astigmatic data.
    ast : float
        Angle for astigmatic data (in degrees).
    ampcon : float
        Amplitude contrast fraction (between 0.0 and 1.0).
    invert_contrast: bool
        Whether to invert the contrast of the input image.
    Cs : float
        Spherical aberration (in mm).
    kV : float
        Acceleration voltage of the TEM (in kV).
    apix : float
        Input pixel size in Angstroms.
    phase_flip :  bool
        Correct CTF by phase-flipping only. This corrects phases but leaves the \
amplitudes unchanged.
    ctf_multiply :  bool
        Correct CTF by multiplying the input image FT with the CTF. This corrects \
phases and dampens amplitudes even further near the CTF zeros.
    wiener_filter : bool
        Correct CTF by applying a Wiener filter. This corrects phases and attempts to \
restore amplitudes to their original values based on an ad-hoc spectral signal-to-noise\
ratio (SSNR) value. That means, at frequencies where the SNR is low amplitudes will be\
restored conservatively, while where the SNR is high these frequencies will be boosted.
    C : float
        Wiener filter constant (per frequency). A scalar can be given to use the same \
constant for all frequencies, whereas an 1D array (radial average) can be given to \
restore each frequency using a different SNR value.
    return_ctf : bool
        Whether to return an array containing the CTF itself alongside the corrected \
image.

    Returns
    -------
    CTFcor : list
        A list containing the corrected image(s), the CTF (if return_ctf is True) and \
a string indicating the type of correction(s) applied.

    Raises
    ------
    ValueError
        If any entry of the Wiener filter constant is lower than or equal to zero.

    """
    if df2 is None:
        df2 = df1

    # Direct CTF correction would invert the image contrast. By default we don't do
    # that, hence the negative sign:
    CTFim = -CTF(img.shape, df1, df2, ast, ampcon, Cs, kV, apix, 0.0, rfft=True)

    if invert_contrast:
        CTFim = -CTFim

        pass

    FT = np.fft.rfftn(img)

    if phase_flip:  # Phase-flipping
        s = np.sign(CTFim)
        CTFcor = np.fft.irfftn(FT * s)

    elif ctf_multiply:  # CTF multiplication
        CTFcor = np.fft.irfftn(FT * CTFim)

    elif wiener_filter:  # Wiener filtering
        if np.any(C <= 0.0):
            raise ValueError(
                "Error: Wiener filter contain value(s) less than or equal to zero!"
            )

        CTFcor = np.fft.irfftn(FT * CTFim / (CTFim * CTFim + C))

    if return_ctf:
        return CTFcor, CTFim

    else:
        return CTFcor


def AdhocSSNR(
    imsize: tuple = (128, 128),
    apix: float = 1.0,
    df: float = 50000.0,
    ampcon: float = 0.07,
    Cs: float = 2.7,
    kV: float = 300.0,
    S: float = 1.0,
    F: float = 1.0,
    hp_frac: float = 0.02,
    lp: bool = True,
):
    """
    An ad hoc SSNR model for cryo-EM data as proposed by Dimitry Tegunov [1,2].

    Parameters
    ----------
    imsize : tuple
        The shape of the input array.
    apix : float
        Input pixel size in Angstroms.
    df : float
        Average defocus in Angstroms.
    ampcon : float
        Amplitude contrast fraction (between 0.0 and 1.0).
    Cs : float
        Spherical aberration (in mm).
    kV : float
        Acceleration voltage of the TEM (in kV).
    S : float
        Strength of the deconvolution to be applied.
    F : float
        Strength of the SSNR falloff.
    hp_frac : float
        fraction of Nyquist frequency to be cut off on the lower end (since it will \
be boosted the most).
    lp : bool
        Whether to low-pass all information beyond the first zero of the CTF.

    Returns
    -------
    ssnr : ndarray
        Array containing the radial ad hoc SSNR.

    Notes
    -----
    This SSNR model ignores astigmatism.

    References
    ----------
    [1] Tegunov & Cramer, Nat. Meth. (2019). https://doi.org/10.1038/s41592-019-0580-y
    [2] https://github.com/dtegunov/tom_deconv/blob/master/tom_deconv.m
    """
    rmesh = RadialIndices(imsize, rounding=False, normalize=True, rfft=True)[0] / apix

    # The ad hoc SSNR exponential falloff
    falloff = np.exp(-100 * rmesh * F) * 10 ** (3 * S)

    # The cosine-shaped high-pass filter. It starts at zero frequency and reaches 1.0
    # at hp_freq (fraction of the Nyquist frequency)
    a = np.minimum(1.0, rmesh * apix / hp_frac)
    highpass = 1.0 - np.cos(a * pi / 2)

    if lp:
        # Ensure the filter will reach zero at the first zero of the CTF
        first_zero_res = FirstZeroCTF(df=df, ampcon=ampcon, Cs=Cs, kV=kV)
        a = np.minimum(1.0, rmesh / first_zero_res)

        lowpass = np.cos(a * pi / 2)

        # Composite filter
        ssnr = highpass * falloff * lowpass

    else:
        ssnr = highpass * falloff  # Composite filter

    return np.abs(ssnr)


def ElectronWavelength(kV: float = 300.0):
    """
    Calculates electron wavelength given acceleration voltage.

    Parameters
    ----------
    kV : float
        Acceleration voltage of the TEM (in kV).

    Returns
    -------
    WL : float
        A scalar value containing the electron wavelength in Angstroms.
    """
    WL = 12.2639 / np.sqrt(kV * 1e3 + 0.97845 * kV * kV)

    return WL


def FirstZeroCTF(
    df: float = 50000.0, ampcon: float = 0.07, Cs: float = 2.7, kV: float = 300.0
):
    """
    The frequency at which the CTF first crosses zero.

    Parameters
    ----------
    df : float
        Average defocus in Angstroms. 
    ampcon : float
        Amplitude contrast fraction (between 0.0 and 1.0).
    Cs : float
        Spherical aberration (in mm).
    kV : float
        Acceleration voltage of the TEM (in kV).

    Returns
    -------
    g : float
        A scalar containing the resolution in Angstroms corresponding to the first \
zero crossing of the CTF.

    Notes
    -----
    Finds the resolution at the first zero of the CTF
    Wolfram Alpha, solving for -w1 * sinXr - w2 * cosXr = 0
    https://www.wolframalpha.com/input/?i=solve+%CF%80*L*(g%5E2)*(d-1%2F(\
    2*(L%5E2)*(g%5E2)*C))%3Dn+%CF%80+-+tan%5E(-1)(c%2Fa)+for+g
    """
    Cs *= 1e7  # Convert Cs to Angstroms

    w1 = np.sqrt(1 - ampcon * ampcon)
    w2 = ampcon

    WL = ElectronWavelength(kV)

    g = np.sqrt(-2 * Cs * WL * np.arctan2(w2, w1) + 2 * pi * Cs * WL + pi) / (
        np.sqrt(2 * pi * Cs * df) * WL
    )

    return g



from typing import Callable, List, Union

import numpy as np
import torch
from monai.transforms import (
    Compose,
    OneOf,
    RandAxisFlipd,
    RandGaussianNoised,
    RandGaussianSmoothd,
    RandRotate90d,
    RandRotated,
    RandZoomd,
    ToTensor,
    ToTensord,
)

from membrain_seg.segmentation.dataloading.transforms import (
    AxesShuffle,
    BlankCuboidTransform,
    BrightnessGradientAdditiveTransform,
    DownsampleSegForDeepSupervisionTransform,
    LocalGammaTransform,
    MedianFilterd,
    RandAdjustContrastWithInversionAndStats,
    RandApplyTransform,
    RandomBrightnessTransformd,
    RandomContrastTransformd,
    SharpeningTransformMONAI,
    SimulateLowResolutionTransform,
)

### Hard-coded area

# Hard-coding kernel sizes for pooling operations (work also with smaller network sizes)
pool_op_kernel_sizes = [
    [2, 2, 2],
    [2, 2, 2],
    [2, 2, 2],
    [2, 2, 2],
    [2, 2, 2],
]  # hard-coded
net_num_pool_op_kernel_sizes = pool_op_kernel_sizes
deep_supervision_scales = [[1, 1, 1]] + [
    list(i) for i in 1 / np.cumprod(np.vstack(net_num_pool_op_kernel_sizes), axis=0)
][:-1]

# Define 3D rotation ranges
data_aug_params = {}
data_aug_params["rotation_x"] = (-30.0 / 360 * 2.0 * np.pi, 30.0 / 360 * 2.0 * np.pi)
data_aug_params["rotation_y"] = (-30.0 / 360 * 2.0 * np.pi, 30.0 / 360 * 2.0 * np.pi)
data_aug_params["rotation_z"] = (-30.0 / 360 * 2.0 * np.pi, 30.0 / 360 * 2.0 * np.pi)

# Which axes should be used for mirroring?
mirror_axes = (0, 1, 2)


def get_mirrored_img(img: torch.Tensor, mirror_idx: int) -> torch.Tensor:
    """
    Get mirrored images for test time augmentation.

    There are 8 possible cases, enumerated from 0 to 7.
    The function supports mirroring across three axes,
    and combinations thereof.

    Parameters
    ----------
    img : torch.Tensor
        Input tensor to be mirrored.
    mirror_idx : int
        Integer index to select mirror case.
        Should be within the range [0, 7] inclusive.

    Returns
    -------
    torch.Tensor
        The mirrored image tensor.

    Raises
    ------
    AssertionError
        If the mirror index is not in the range [0, 7].

    """
    assert mirror_idx < 8 and mirror_idx >= 0
    if mirror_idx == 0:
        return img

    if mirror_idx == 1 and (2 in mirror_axes):
        return torch.flip(img, (4,))

    if mirror_idx == 2 and (1 in mirror_axes):
        return torch.flip(img, (3,))

    if mirror_idx == 3 and (2 in mirror_axes) and (1 in mirror_axes):
        return torch.flip(img, (4, 3))

    if mirror_idx == 4 and (0 in mirror_axes):
        return torch.flip(img, (2,))

    if mirror_idx == 5 and (0 in mirror_axes) and (2 in mirror_axes):
        return torch.flip(img, (4, 2))

    if mirror_idx == 6 and (0 in mirror_axes) and (1 in mirror_axes):
        return torch.flip(img, (3, 2))

    if (
        mirror_idx == 7
        and (0 in mirror_axes)
        and (1 in mirror_axes)
        and (2 in mirror_axes)
    ):
        return torch.flip(img, (4, 3, 2))


def get_training_transforms(
    prob_to_one: bool = False, return_as_list: bool = False
) -> Union[List[Callable], Compose]:
    """
    Returns the data augmentation transforms for training phase.

    The function sets up an augmentation sequence containing a variety of
    transformations such as rotations, zooms, 90 degree rotations,
    Gaussian noise, brightness adjustments, and more.
    If desired, the sequence can be returned as a list.

    Parameters
    ----------
    prob_to_one : bool, optional
        If True, the probability of applying the transformation is set to 1.0 for
            all transformations in the sequence.
        If False, the probability is lower (specified within each transformation).
        Default is False.
    return_as_list : bool, optional
        If True, the sequence of transformations is returned as a list.
        If False, the sequence is returned as a Compose object. Default is False.

    Returns
    -------
    List[Callable] or Compose
        If return_as_list is True, the function returns a list of
            transformation functions.
        If return_as_list is False, the function returns a Compose object
            containing the sequence of transformations.

    """
    aug_sequence = [
        RandRotated(
            keys=("image", "label"),
            range_x=data_aug_params["rotation_x"],
            range_y=data_aug_params["rotation_y"],
            range_z=data_aug_params["rotation_x"],
            prob=(1.0 if prob_to_one else 0.75),
            mode=("bilinear", "nearest"),
        ),
        RandZoomd(
            keys=("image", "label"),
            prob=(1.0 if prob_to_one else 0.3),
            min_zoom=0.7,
            max_zoom=1.43,
            mode=("trilinear", "nearest-exact"),
            padding_mode=("constant", "constant"),
        ),  # TODO: Independent scale for each axis?
        RandRotate90d(
            keys=("image", "label"),
            prob=(1.0 if prob_to_one else 0.70),
            max_k=3,
            spatial_axes=(0, 1),
        ),
        RandRotate90d(
            keys=("image", "label"),
            prob=(1.0 if prob_to_one else 0.70),
            max_k=3,
            spatial_axes=(0, 2),
        ),
        RandRotate90d(
            keys=("image", "label"),
            prob=(1.0 if prob_to_one else 0.70),
            max_k=3,
            spatial_axes=(1, 2),
        ),
        AxesShuffle,
        OneOf(
            [
                RandApplyTransform(
                    transform=MedianFilterd(keys=["image"], radius=(2, 8)),
                    prob=(
                        1.0 if prob_to_one else 0.25
                    ),  # Changed range from 8 to 6 and prob to 15%
                    # for efficiency reasons
                ),
                RandGaussianSmoothd(
                    keys=["image"],
                    sigma_x=(0.3, 1.5),
                    sigma_y=(0.3, 1.5),
                    sigma_z=(0.3, 1.5),
                    prob=(1.0 if prob_to_one else 0.3),
                ),
            ]
        ),
        RandGaussianNoised(
            keys=["image"], prob=(1.0 if prob_to_one else 0.4), mean=0.0, std=0.7
        ),  # Chaned std from 0.1 to 0.5 --> check visually
        RandomBrightnessTransformd(
            keys=["image"], mu=0.0, sigma=0.5, prob=(1.0 if prob_to_one else 0.30)
        ),
        OneOf(
            [
                RandomContrastTransformd(
                    keys=["image"],
                    contrast_range=(0.5, 2.0),
                    preserve_range=True,
                    prob=(1.0 if prob_to_one else 0.30),
                ),
                RandomContrastTransformd(
                    keys=["image"],
                    contrast_range=(0.5, 2.0),
                    preserve_range=False,
                    prob=(1.0 if prob_to_one else 0.30),
                ),
            ]
        ),
        RandApplyTransform(
            SimulateLowResolutionTransform(
                keys=["image"],
                downscale_factor_range=(0.25, 1.0),
                upscale_mode="trilinear",
                downscale_mode="nearest-exact",
            ),
            prob=(1.0 if prob_to_one else 0.35),
        ),
        RandApplyTransform(
            Compose(
                [
                    RandAdjustContrastWithInversionAndStats(keys=["image"], prob=1.0),
                    RandAdjustContrastWithInversionAndStats(keys=["image"], prob=1.0),
                ]
            ),
            prob=(1.0 if prob_to_one else 0.25),
        ),
        RandAxisFlipd(keys=("image", "label"), prob=(0.5)),
        BlankCuboidTransform(
            keys=["image"],
            prob=(1.0 if prob_to_one else 0.4),
            cuboid_area=(160 // 10, 160 // 3),
            is_3d=True,
            max_cuboids=5,
            replace_with="mean",
        ),  # patch size of 160 hard-coded. Should we make it flexible?
        RandApplyTransform(
            BrightnessGradientAdditiveTransform(
                keys=["image"],
                scale=lambda x, y: np.exp(
                    np.random.uniform(np.log(x[y] // 6), np.log(x[y]))
                ),
                loc=(-0.5, 1.5),
                max_strength=lambda x, y: (
                    np.random.uniform(-5, -1)
                    if np.random.uniform() < 0.5
                    else np.random.uniform(1, 5)
                ),
                mean_centered=False,
            ),
            prob=(1.0 if prob_to_one else 0.3),
        ),
        RandApplyTransform(
            LocalGammaTransform(
                keys=["image"],
                scale=lambda x, y: np.exp(
                    np.random.uniform(np.log(x[y] // 6), np.log(x[y]))
                ),
                loc=(-0.5, 1.5),
                gamma=lambda: (
                    np.random.uniform(0.01, 0.8)
                    if np.random.uniform() < 0.5
                    else np.random.uniform(1.5, 4)
                ),
            ),
            prob=(1.0 if prob_to_one else 0.3),
        ),
        SharpeningTransformMONAI(
            keys=["image"],
            strength=(0.1, 1),
            same_for_each_channel=False,
            p_per_channel=(1.0 if prob_to_one else 0.2),
        ),
        DownsampleSegForDeepSupervisionTransform(
            keys=["label"], ds_scales=deep_supervision_scales, order="nearest"
        ),
        ToTensord(keys=["image"], dtype=torch.float),
    ]
    if return_as_list:
        return aug_sequence
    return Compose(aug_sequence)


def get_validation_transforms(
    return_as_list: bool = False,
) -> Union[List[Callable], Compose]:
    """
    Returns the data augmentation transforms for the validation phase.

    The function sets up a sequence of transformations including downsampling
    and tensor conversion. If desired, the sequence can be returned as a list.

    Parameters
    ----------
    return_as_list : bool, optional
        If True, the sequence of transformations is returned as a list.
        If False, the sequence is returned as a Compose object. Default is False.

    Returns
    -------
    List[Callable] or Compose
        If return_as_list is True, the function returns a list of
            transformation functions.
        If return_as_list is False, the function returns a Compose object
            containing the sequence of transformations.

    """
    aug_sequence = [
        DownsampleSegForDeepSupervisionTransform(
            keys=["label"], ds_scales=deep_supervision_scales, order="nearest"
        ),
        ToTensord(keys=["image"], dtype=torch.float),
    ]
    if return_as_list:
        return aug_sequence
    return Compose(aug_sequence)


def get_prediction_transforms() -> Compose:
    """
    Returns the data augmentation transforms for the prediction phase.

    The function sets up a Compose object containing a transformation for
    converting data to tensors.

    Returns
    -------
    Compose
        A Compose object containing the sequence of transformations for
        the prediction phase.

    """
    transforms = Compose(
        [
            ToTensor(),
        ]
    )
    return transforms



# ---------------------------------------------------------------------------------
# DISCLAIMER: This code is adapted from the MATLAB and C++ implementations provided
# in the paper titled "Robust membrane detection based on tensor voting for electron
# tomography" by Antonio Martinez-Sanchez, Inmaculada Garcia, Shoh Asano, Vladan Lucic,
# and Jose-Jesus Fernandez, published in the Journal of Structural Biology,
# Volume 186, Issue 1, 2014, Pages 49-61. The original work can be accessed via
# https://doi.org/10.1016/j.jsb.2014.02.015 and is used under conditions that adhere
# to the original licensing agreements. For details on the original license, refer to
# the publication: https://www.sciencedirect.com/science/article/pii/S1047847714000495.
# ---------------------------------------------------------------------------------
import logging

import numpy as np
import scipy.ndimage as ndimage
import torch
from skimage.morphology import skeletonize

from membrain_seg.segmentation.skeletonization.diff3d import (
    compute_gradients,
    compute_hessian,
)
from membrain_seg.segmentation.skeletonization.eig3d import (
    batch_mask_eigendecomposition_3d,
)
from membrain_seg.segmentation.skeletonization.nonmaxsup import nonmaxsup
from membrain_seg.segmentation.training.surface_dice import apply_gaussian_filter


def skeletonization(segmentation: np.ndarray, batch_size: int, device: str=None) -> np.ndarray:
    """
    Perform skeletonization on a tomogram segmentation.

    This function skeletonizes the input segmentation where the non-zero labels
    represent the structures of interest. The resultan skeleton is saved with
    '_skel' appended after the filename.

    Parameters
    ----------
    segmentation : ndarray
        Tomogram segmentation as a numpy array, where non-zero values represent
        the structures of interest.

    batch_size : int
        The number of elements to process in one batch during eigen decomposition.
        Useful for managing memory usage.

    Returns
    -------
    ndarray
        Returns the skeletonized image as a numpy array.

    Notes
    -----
    The skeletonization is based on the computation of the distance transform
    of the non-zero regions (foreground), followed by an eigenvalue analysis
    of the Hessian matrix of the distance transform to identify ridge-like
    structures corresponding to the centerlines of the segmented objects.

    Examples
    --------
    >>> membrain skeletonize --label-path <path> --out-folder <output-directory>
        --batch-size 1000000
    This command runs the skeletonization process from the command line.
    """
    # Convert non-zero segmentation values to 1.0
    labels = (segmentation > 0) * 1.0

    labels_dt = ndimage.distance_transform_edt(labels) * (-1)

    # Calculates partial derivative along 3 dimensions.
    gradX, gradY, gradZ = compute_gradients(labels_dt)

    # Calculates Hessian tensor
    hessianXX, hessianYY, hessianZZ, hessianXY, hessianXZ, hessianYZ = compute_hessian(
        gradX, gradY, gradZ
    )
    hessians = [hessianXX, hessianYY, hessianZZ, hessianXY, hessianXZ, hessianYZ]
    del gradX, gradY, gradZ

    # Apply Gaussian filter with the same sigma value for all dimensions
    # Load hessian tensors on GPU
    if device is None:
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
    else:
        assert device in ["cuda", "cpu"], "Device must be either 'cuda' or 'cpu'."
        device = torch.device(device)

    filtered_hessian = [
        apply_gaussian_filter(
            torch.from_numpy(comp).float().to(device).unsqueeze(0).unsqueeze(0),
            kernel_size=9,
            sigma=1.0,
        )
        .squeeze()
        .to("cpu")
        for comp in hessians
    ]

    # Solve Eigen problem
    logging.info("Computing Eigenvalues and Eigenvectors.")
    logging.info(
        "In case the execution of the program is terminated unexpectedly, "
        "attempt to rerun it using smaller segmentation patches"
        "or give a specified batch size as input, e.g. batch_size=1000000."
    )
    first_eigenvalue, first_eigenvector = batch_mask_eigendecomposition_3d(
        filtered_hessian, batch_size, labels
    )

    # Non-maximum suppression
    first_eigenvalue = ndimage.gaussian_filter(first_eigenvalue, sigma=1)
    skeleton = nonmaxsup(
        first_eigenvalue,
        first_eigenvector[:, :, :, 0],
        first_eigenvector[:, :, :, 1],
        first_eigenvector[:, :, :, 2],
        labels,
    )
    return skeleton


def skeletonize_skimage(segmentation: np.ndarray) -> np.ndarray:
    """
    Perform skeletonization using skimage's skeletonize function.

    Skeletonization is performed slice by slice on the input segmentation where
    the non-zero labels represent the structures of interest. The resulting
    skeleton is saved with '_skel' appended after the filename.

    Parameters
    ----------
    segmentation : ndarray
        Tomogram segmentation as a numpy array, where non-zero values represent
        the structures of interest.

    Returns
    -------
    ndarray
        Returns the skeletonized image as a numpy array.
    """
    # Convert non-zero segmentation values to 1
    binary_seg = (segmentation > 0).astype(np.uint8)
    new_seg = np.zeros_like(binary_seg)

    for i in range(binary_seg.shape[2]):
        slice_skel = binary_seg[:, :, i]
        if np.sum(slice_skel) > 0:
            # make contiguous
            slice_skel = np.ascontiguousarray(slice_skel)
            # skeletonize
            slice_skel = skeletonize(slice_skel)
            new_seg[:, :, i] = slice_skel.astype(np.uint8)
    return new_seg



import logging
import os
from typing import List

from typer import Option

from .cli import OPTION_PROMPT_KWARGS as PKWARGS
from .cli import cli

logging.basicConfig(level=logging.INFO)


@cli.command(name="segment", no_args_is_help=True)
def segment(
    tomogram_path: str = Option(  # noqa: B008
        ..., help="Path to the tomogram to be segmented", **PKWARGS
    ),
    ckpt_path: str = Option(  # noqa: B008
        ...,
        help="Path to the pre-trained model checkpoint that should be used.",
        **PKWARGS,
    ),
    out_folder: str = Option(  # noqa: B008
        "./predictions", help="Path to the folder where segmentations should be stored."
    ),
    rescale_patches: bool = Option(  # noqa: B008
        False, help="Should patches be rescaled on-the-fly during inference?"
    ),
    in_pixel_size: float = Option(  # noqa: B008
        None,
        help="Pixel size of the input tomogram in Angstrom. \
            (default: 10 Angstrom)",
    ),
    out_pixel_size: float = Option(  # noqa: B008
        10.0,
        help="Pixel size of the output segmentation in Angstrom. \
            (default: 10 Angstrom; should normally stay at 10 Angstrom)",
    ),
    store_probabilities: bool = Option(  # noqa: B008
        False, help="Should probability maps be output in addition to segmentations?"
    ),
    store_connected_components: bool = Option(  # noqa: B008
        False, help="Should connected components of the segmentation be computed?"
    ),
    connected_component_thres: int = Option(  # noqa: B008
        None,
        help="Threshold for connected components. Components smaller than this will \
            be removed from the segmentation.",
    ),
    test_time_augmentation: bool = Option(  # noqa: B008
        True,
        help="Use 8-fold test time augmentation (TTA)? TTA improves segmentation \
        quality slightly, but also increases runtime.",
    ),
    store_uncertainty_map: bool = Option(
        False,
        help="If True, store an uncertainty map based on the voxel-wise variance "
            "across the TTA predictions. Requires --test-time-augmentation=True.",
    ),
    segmentation_threshold: float = Option(  # noqa: B008
        0.0,
        help="Threshold for the membrane segmentation. Only voxels with a \
            membrane score higher than this threshold will be segmented. \
                (default: 0.0)",
    ),
    sliding_window_size: int = Option(  # noqa: B008
        160,
        help="Sliding window size used for inference. Smaller values than 160 \
            consume less GPU, but also lead to worse segmentation results!",
    ),
):
    """Segment tomograms using a trained model.

    Example
    -------
    membrain segment --tomogram-path <path-to-your-tomo>
    --ckpt-path <path-to-your-model>
    """
    from membrain_seg.segmentation.segment import segment as _segment

    print("Segmenting tomogram", tomogram_path)
    print("")
    print(
        "This can take several minutes. If you are bored, why not learn about \
            what's happening under the hood by reading the MemBrain v2 preprint?"
    )
    print(
        "MemBrain v2: an end-to-end tool for the analysis of membranes in \
            cryo-electron tomography"
    )
    print("https://www.biorxiv.org/content/10.1101/2024.01.05.574336v1")
    print("")
    _segment(
        tomogram_path=tomogram_path,
        ckpt_path=ckpt_path,
        out_folder=out_folder,
        rescale_patches=rescale_patches,
        in_pixel_size=in_pixel_size,
        out_pixel_size=out_pixel_size,
        store_probabilities=store_probabilities,
        store_connected_components=store_connected_components,
        connected_component_thres=connected_component_thres,
        sw_roi_size=sliding_window_size,
        test_time_augmentation=test_time_augmentation,
        store_uncertainty_map=store_uncertainty_map,
        segmentation_threshold=segmentation_threshold,
    )


@cli.command(name="components", no_args_is_help=True)
def components(
    segmentation_path: str = Option(  # noqa: B008
        ..., help="Path to the membrane segmentation to be processed.", **PKWARGS
    ),
    out_folder: str = Option(  # noqa: B008
        "./predictions", help="Path to the folder where segmentations should be stored."
    ),
    connected_component_thres: int = Option(  # noqa: B008
        None,
        help="Threshold for connected components. Components smaller than this will \
            be removed from the segmentation.",
    ),
):
    """Compute connected components of your segmented tomogram.

    This will annotate the connected components of your segmentation
    with integer values from 1 to [number of membranes].

    Example
    -------
    membrain components --tomogram-path <path-to-your-tomo>
    --connected-component-thres 5
    """
    from membrain_seg.segmentation.connected_components import (
        connected_components as _connected_components,
    )
    from membrain_seg.segmentation.dataloading.data_utils import (
        load_tomogram,
        store_tomogram,
    )

    segmentation = load_tomogram(segmentation_path)
    conn_comps = _connected_components(
        binary_seg=segmentation.data, size_thres=connected_component_thres
    )
    segmentation.data = conn_comps
    out_file = os.path.join(
        out_folder,
        os.path.splitext(os.path.basename(segmentation_path))[0] + "_components.mrc",
    )
    store_tomogram(filename=out_file, tomogram=segmentation)


@cli.command(name="thresholds", no_args_is_help=True)
def thresholds(
    scoremap_path: str = Option(  # noqa: B008
        ..., help="Path to the membrane scoremap to be processed.", **PKWARGS
    ),
    thresholds: List[float] = Option(  # noqa: B008
        ...,
        help="List of thresholds. Provide multiple by repeating the option.",
    ),
    out_folder: str = Option(  # noqa: B008
        "./predictions",
        help="Path to the folder where thresholdedsegmentations \
            should be stored.",
    ),
):
    """Process the provided scoremap using given thresholds.

    Given a membrane scoremap, this function thresholds the scoremap data
    using the provided threshold(s). The thresholded scoremaps are then stored
    in the specified output folder. If multiple thresholds are provided,
    separate thresholded scoremaps will be generated for each threshold.

    Example
    -------
    membrain thresholds --scoremap-path <path-to-scoremap>
        --thresholds -1.5 --thresholds -0.5 --thresholds 0.0 --thresholds 0.5

    This will generate thresholded scoremaps for the provided scoremap at
    thresholds -1.5, -0.5, 0.0 and 0.5.The results will be saved with filenames
    indicating the threshold values in the default 'predictions' folder or
    in the folder specified by the user.
    """
    from membrain_seg.segmentation.dataloading.data_utils import (
        load_tomogram,
        store_tomogram,
    )

    scoremap = load_tomogram(scoremap_path)
    score_data = scoremap.data
    if not isinstance(thresholds, list):
        thresholds = [thresholds]
    for threshold in thresholds:
        logging.info("Thresholding at" + str(threshold))
        thresholded_data = score_data > threshold
        segmentation = scoremap
        segmentation.data = thresholded_data
        out_file = os.path.join(
            out_folder,
            os.path.splitext(os.path.basename(scoremap_path))[0]
            + f"_threshold_{threshold}.mrc",
        )
        store_tomogram(filename=out_file, tomogram=segmentation)
        logging.info("Saved thresholded scoremap to " + out_file)



import logging

import pytorch_lightning as pl
from pytorch_lightning import loggers as pl_loggers
from pytorch_lightning.callbacks import LearningRateMonitor, ModelCheckpoint

from membrain_seg.segmentation.dataloading.memseg_pl_datamodule import (
    MemBrainSegDataModule,
)
from membrain_seg.segmentation.networks.unet import SemanticSegmentationUnet
from membrain_seg.segmentation.training.optim_utils import (
    ToleranceCallback,
)
from membrain_seg.segmentation.training.training_param_summary import (
    print_training_parameters,
)


def fine_tune(
    pretrained_checkpoint_path: str,
    finetune_data_dir: str,
    finetune_learning_rate: float = 1e-5,
    log_dir: str = "logs_finetune/",
    batch_size: int = 2,
    num_workers: int = 8,
    max_epochs: int = 100,
    early_stop_threshold: float = 0.05,
    aug_prob_to_one: bool = False,
    use_deep_supervision: bool = False,
    project_name: str = "membrain-seg_finetune",
    sub_name: str = "1",
    use_surf_dice: bool = False,
    surf_dice_weight: float = 1.0,
    surf_dice_tokens: list = None,
) -> None:
    """
    Fine-tune a pre-trained U-Net model on new datasets.

    This function finetunes a pre-trained U-Net model on new data provided by the user.
    The `finetune_data_dir` should contain the following directories:
    - `imagesTr` and `labelsTr` for the user's own new training data.
    - `imagesVal` and `labelsVal` for the old data, which will be used
      for validation to ensure that the fine-tuned model's performance
      is not significantly worse on the original training data than the
      pre-trained model.

    Callbacks used during the fine-tuning process
    ---------
    - ModelCheckpoint: Saves the model checkpoints based on training loss
      and at regular intervals.
    - ToleranceCallback: Stops training if the validation loss deviates significantly
      from the baseline value set after the first epoch.
    - LearningRateMonitor: Monitors and logs the learning rate during training.
    - PrintLearningRate: Prints the current learning rate at the start of each epoch.

    Parameters
    ----------
    pretrained_checkpoint_path : str
        Path to the checkpoint of the pre-trained model.
    finetune_data_dir : str
        Path to the directory containing the new data for fine-tuning
        and old data for validation.
    finetune_learning_rate : float, optional
        Learning rate for fine-tuning, by default 1e-5.
    log_dir : str, optional
        Path to the directory where logs should be stored.
    batch_size : int, optional
        Number of samples per batch of input data.
    num_workers : int, optional
        Number of subprocesses to use for data loading.
    max_epochs : int, optional
        Maximum number of epochs to finetune, by default 100.
    early_stop_threshold : float, optional
        Threshold for early stopping based on validation loss deviation,
        by default 0.05.
    aug_prob_to_one : bool, optional
        If True, all augmentation probabilities are set to 1.
    use_deep_supervision : bool, optional
        If True, enables deep supervision in the U-Net model.
    project_name : str, optional
        Name of the project for logging purposes.
    sub_name : str, optional
        Sub-name of the project for logging purposes.
    use_surf_dice : bool, optional
        If True, enables Surface-Dice loss.
    surf_dice_weight : float, optional
        Weight for the Surface-Dice loss.
    surf_dice_tokens : list, optional
        List of tokens to use for the Surface-Dice loss.

    Returns
    -------
    None
    """
    # Print training parameters for verification
    print_training_parameters(
        data_dir=finetune_data_dir,
        log_dir=log_dir,
        batch_size=batch_size,
        num_workers=num_workers,
        max_epochs=max_epochs,
        aug_prob_to_one=aug_prob_to_one,
        use_deep_supervision=use_deep_supervision,
        project_name=project_name,
        sub_name=sub_name,
        use_surf_dice=use_surf_dice,
        surf_dice_weight=surf_dice_weight,
        surf_dice_tokens=surf_dice_tokens,
    )
    logging.info("————————————————————————————————————————————————————————")
    logging.info(
        f"Pretrained Checkpoint:\n"
        f"   '{pretrained_checkpoint_path}' \n"
        f"   Path to the pretrained model checkpoint."
    )
    logging.info("\n")

    # Initialize the data module with fine-tuning datasets
    # New data for finetuning and old data for validation
    finetune_data_module = MemBrainSegDataModule(
        data_dir=finetune_data_dir,
        batch_size=batch_size,
        num_workers=num_workers,
        aug_prob_to_one=aug_prob_to_one,
    )

    # Load the pre-trained model with updated learning rate
    pretrained_model = SemanticSegmentationUnet.load_from_checkpoint(
        pretrained_checkpoint_path, learning_rate=finetune_learning_rate
    )

    checkpointing_name = project_name + "_" + sub_name

    # Set up logging
    csv_logger = pl_loggers.CSVLogger(log_dir)

    # Set up model checkpointing based on training loss
    checkpoint_callback_train_loss = ModelCheckpoint(
        dirpath="finetuned_checkpoints/",
        filename=checkpointing_name + "-{epoch:02d}-{train_loss:.2f}",
        monitor="train_loss",
        mode="min",
        save_top_k=3,
    )

    # Set up regular checkpointing every 5 epochs
    checkpoint_callback_regular = ModelCheckpoint(
        save_top_k=-1,  # Save all checkpoints
        every_n_epochs=5,
        dirpath="finetuned_checkpoints/",
        filename=checkpointing_name + "-{epoch}-{train_loss:.2f}",
        verbose=True,  # Print a message when a checkpoint is saved
    )

    # Set up ToleranceCallback by monitoring validation loss
    early_stop_metric = "val_loss"
    tolerance_callback = ToleranceCallback(early_stop_metric, early_stop_threshold)

    # Monitor learning rate changes
    lr_monitor = LearningRateMonitor(logging_interval="epoch", log_momentum=True)

    # Initialize the trainer with specified precision, logger, and callbacks
    trainer = pl.Trainer(
        precision="16-mixed",
        logger=[csv_logger],
        callbacks=[
            checkpoint_callback_train_loss,
            checkpoint_callback_regular,
            lr_monitor,
            tolerance_callback,
        ],
        max_epochs=max_epochs,
    )

    # Start the fine-tuning process
    trainer.fit(pretrained_model, finetune_data_module)



from typer import Option

from ..cli import OPTION_PROMPT_KWARGS as PKWARGS
from ..cli import cli
from .match_pixel_size import match_pixel_size as _match_pixel_size
from .match_pixel_size_seg import match_segmentation_pixel_size_to_tomo


@cli.command(name="match_pixel_size", no_args_is_help=True)
def match_pixel_size(
    input_tomogram: str = Option(  # noqa: B008
        ..., help="Path to the input tomogram", **PKWARGS
    ),
    output_path: str = Option(  # noqa: B008
        ..., help="Path to store the output files", **PKWARGS
    ),
    pixel_size_out: float = Option(10.0, help="Target pixel size."),  # noqa: B008
    pixel_size_in: float = Option(  # noqa: B008
        None,
        help="Input pixel size (optional). If not \
            specified, it will be read from the tomogram's header. ATTENTION: This can \
            lead to severe errors if the header pixel size is not correct.",
    ),
    disable_smooth: bool = Option(  # noqa: B008
        False,
        help="Disable smoothing (ellipsoid mask + \
            cosine decay). Disable if causing problems or for speed up",
    ),
):
    """Tomogram pixel size matching.

    Resize your tomogram to the specified target pixel size. We recommend
    to use MemBrain-seg with a pixel size of around 10A/pixel.
    """
    _match_pixel_size(
        input_tomogram, output_path, pixel_size_in, pixel_size_out, disable_smooth
    )


@cli.command(name="match_seg_to_tomo", no_args_is_help=True)
def match_seg_to_tomo(
    seg_path: str = Option(  # noqa: B008
        ..., help="Path to the segmentation", **PKWARGS
    ),
    orig_tomo_path: str = Option(  # noqa: B008
        ...,
        help="Path to the tomogram the segmentation should be matched to.",
        **PKWARGS,
    ),
    output_path: str = Option(  # noqa: B008
        ...,
        help="Path to the where the resized segmentation should be stored.",
        **PKWARGS,
    ),
):
    """
    Segmentation pixel size matching.

    Using this function, you can resize a segmentation (binary / multi-label) to
    the size of a specified tomogram.
    Use case: You have matched your input tomogram's pixel size to the training
    pixel size (around 10A), have performed MemBrain-seg's segmentation, and now
    would like to have the segmentation in the scale of the original tomogram.
    """
    # Call your function here
    match_segmentation_pixel_size_to_tomo(seg_path, orig_tomo_path, output_path)



import numpy as np
from scipy import ndimage


def connected_components(binary_seg: np.ndarray, size_thres: int = None):
    """
    Compute connected components from a 3D membrane segmentation.

    The function uses 3D connectivity for label assignments. If a size
    threshold is specified, components having size smaller than the
    threshold will be removed.


    Parameters
    ----------
    binary_seg : np.ndarray
        Input 3D binary image with shape (X, Y, Z).
        Non-zero elements are considered as the foreground.

    size_thres : int, optional
        The size threshold for removing small connected components.
        Components having size (number of voxels) smaller than this
        value will be removed.
        If None (default), no components are removed.

    Returns
    -------
    labeled_array : np.ndarray
        A 3D image where each voxel in a connected component has a
        unique label. The labels are in the range 1 to N, where N is
        the number of connected components found. All background voxels
        are zero.
    """
    structure = np.ones((3, 3, 3), dtype=bool)
    labeled_array, num_features = ndimage.label(binary_seg, structure=structure)
    print(f"Found {num_features} connected components.")
    if size_thres is not None and size_thres > 1:
        sizes = np.bincount(labeled_array.ravel())
        print(f"Removing components smaller than {size_thres} voxels.")
        # keep only components above threshold
        keep_mask = np.isin(labeled_array, np.where(sizes >= size_thres)[0])
        labeled_array = labeled_array * keep_mask
        # relabel to make them contiguous 1..N
        labeled_array, _, _ = relabel_sequential(labeled_array)

    return labeled_array


def relabel_sequential(labeled_array):
    """Make labels contiguous 1..N."""
    unique = np.unique(labeled_array)
    unique = unique[unique != 0]  # skip background
    mapping = {old: new for new, old in enumerate(unique, 1)}
    relabeled = np.zeros_like(labeled_array)
    print(f"Relabeled to {len(unique)} connected components.")
    for old, new in mapping.items():
        relabeled[labeled_array == old] = new
    return relabeled, mapping, len(unique)



#!/usr/bin/env python3
"""
Parallel downloader for large Zenodo files (no external dependencies).

- Uses HTTP Range requests
- Splits file into N chunks and downloads them in parallel
- Merges chunks into final file
- Optional MD5 verification
"""

import hashlib
import math
import os
import sys
import threading
import time
import urllib.request

# --- MemBrain-seg Training data URLs ---------------------------------------

RECORD_ID = "15089686"
FILENAME = "MemBrain_seg_training_data.zip"
DOWNLOAD_URL = f"https://zenodo.org/records/{RECORD_ID}/files/{FILENAME}?download=1"

# Set to None if you don't care about checksum
MD5_EXPECTED = "4553547a60fd5afbda547f54d034ad28"

NUM_WORKERS = 8  # Increase if you have good bandwidth/CPU
CHUNK_SIZE = 1024 * 1024  # 1 MB per read

# ---------------------------------------------------------------------------


def get_filesize(url: str) -> int:
    """Ask server for Content-Length via HEAD."""
    req = urllib.request.Request(url, method="HEAD")
    with urllib.request.urlopen(req) as resp:
        size = resp.getheader("Content-Length")
        if size is None:
            raise RuntimeError("Server did not provide Content-Length header.")
        return int(size)


def download_range(
    url: str,
    start: int,
    end: int,
    part_path: str,
    progress: dict,
    lock: threading.Lock,
    retries: int = 3,
):
    """Download byte range [start, end] to part_path."""
    headers = {"Range": f"bytes={start}-{end}"}
    for attempt in range(retries):
        try:
            req = urllib.request.Request(url, headers=headers)
            with urllib.request.urlopen(req) as resp, open(part_path, "wb") as f:
                while True:
                    chunk = resp.read(CHUNK_SIZE)
                    if not chunk:
                        break
                    f.write(chunk)
                    with lock:
                        progress["downloaded"] += len(chunk)
            return
        except Exception:
            if attempt == retries - 1:
                raise
            # simple exponential backoff
            time.sleep(2**attempt)


def print_progress(progress: dict, total_size: int, lock: threading.Lock):
    """Simple console progress bar."""
    while True:
        with lock:
            downloaded = progress["downloaded"]
            done = progress["done"]
        percent = downloaded / total_size * 100
        bar_len = 40
        filled = int(bar_len * percent / 100)
        bar = "#" * filled + "." * (bar_len - filled)
        sys.stdout.write(
            f"\r[{bar}] {percent:6.2f}%  "
            f"{downloaded / (1024**2):7.2f} / {total_size / (1024**2):.2f} MB"
        )
        sys.stdout.flush()
        if done:
            break
        time.sleep(0.5)
    print()  # newline


def merge_parts(part_paths, out_path):
    """Concatenate part files into final file."""
    with open(out_path, "wb") as outfile:
        for part in part_paths:
            with open(part, "rb") as f:
                while True:
                    chunk = f.read(CHUNK_SIZE)
                    if not chunk:
                        break
                    outfile.write(chunk)


def md5sum(path: str) -> str:
    """Calculate MD5 checksum of a file."""
    h = hashlib.md5()
    with open(path, "rb") as f:
        for chunk in iter(lambda: f.read(1024 * 1024), b""):
            h.update(chunk)
    return h.hexdigest()


def download_file_single_process(out_path, chunk_size=1024 * 1024):
    """Download file using a single process.

    Parameters
    ----------
    out_path : str
        Path to save the downloaded file.
    chunk_size : int
        Number of bytes to read per chunk.
    """
    print(f"Downloading to {out_path}")
    with urllib.request.urlopen(DOWNLOAD_URL) as response:
        total = response.length
        downloaded = 0

        with open(out_path, "wb") as f:
            while True:
                chunk = response.read(chunk_size)
                if not chunk:
                    break
                f.write(chunk)

                downloaded += len(chunk)
                if total:
                    done = int(50 * downloaded / total)
                    bar = "[" + "#" * done + "." * (50 - done) + "]"
                    print(
                        f"\r{bar} "
                        f"{downloaded / (1024**2):.2f} / {total / (1024**2):.2f} MB",
                        end="",
                        flush=True,
                    )
    print("\nDownload finished.")


def download_file_parallel(out_path: str, num_workers: int = 2):
    """
    Download file in parallel using multiple threads.

    Parameters
    ----------
    out_path : str
        Path to save the downloaded file.
    num_workers : int
        Number of parallel download threads to use.
    """
    print("Querying file size from Zenodo…")
    total_size = get_filesize(DOWNLOAD_URL)
    print(f"Total size: {total_size / (1024**3):.2f} GB")

    part_size = math.ceil(total_size / num_workers)
    part_paths = [f"{out_path}.part{i}" for i in range(num_workers)]

    progress = {"downloaded": 0, "done": False}
    lock = threading.Lock()

    # Start progress thread
    prog_thread = threading.Thread(
        target=print_progress,
        args=(progress, total_size, lock),
        daemon=True,
    )
    prog_thread.start()

    # Start worker threads
    threads = []
    for i in range(num_workers):
        start = i * part_size
        end = min(total_size - 1, (i + 1) * part_size - 1)
        if start > end:
            break  # in case num_workers > needed
        t = threading.Thread(
            target=download_range,
            args=(DOWNLOAD_URL, start, end, part_paths[i], progress, lock),
        )
        threads.append(t)
        t.start()

    # Wait for workers
    for t in threads:
        t.join()

    # Flag progress as done
    with lock:
        progress["done"] = True
    prog_thread.join()

    print("Merging parts…")
    merge_parts(part_paths, out_path)

    # Cleanup part files
    for p in part_paths:
        if os.path.exists(p):
            os.remove(p)

    print("Download finished.")


def download_data(out_folder, num_parallel_processes=1):
    """Download the MemBrain-seg dataset from Zenodo.

    Parameters
    ----------
    out_folder : str
        Folder to store the downloaded MemBrain-seg dataset (14GB .zip file).
    num_parallel_processes : int
        Number of parallel processes to use for downloading the dataset.
        Since dataset is hosted on Zenodo with limited bandwidth,
        using multiple parallel processes can speed up the download.
        Pass an integer value greater than 0. For example, to use 4 parallel
        processes, pass "4". Default is "1" (single process).
    """
    out_path = os.path.join(out_folder, FILENAME)
    out_path = FILENAME

    if os.path.exists(out_path):
        print(f"{out_path} already exists, skipping download.")
    elif num_parallel_processes <= 1:
        print("Downloading using single process…")
        print(
            "Note: Download via Zenodo may be slow due to server limits. \
                When in doubt, better get a coffee."
        )
        download_file_single_process(out_path)
    else:
        print(f"Downloading using {num_parallel_processes} parallel processes…")
        print(
            "Note: Download via Zenodo may be slow due to server limits. \
                When in doubt, better get a coffee."
        )
        download_file_parallel(out_path, num_workers=num_parallel_processes)

    # Optional MD5 verification
    if MD5_EXPECTED:
        print("\nVerifying MD5 checksum…")
        md5_actual = md5sum(out_path)
        print(f"Expected: {MD5_EXPECTED}")
        print(f"Actual  : {md5_actual}")
        if md5_actual.lower() == MD5_EXPECTED.lower():
            print("✅ MD5 checksum matches. File OK.")
        else:
            print("❌ MD5 checksum does NOT match! File may be corrupted.")



import logging
from functools import partial
from typing import Dict, Tuple

import pytorch_lightning as pl
import torch
from monai.data import decollate_batch
from monai.metrics import DiceMetric
from monai.transforms import AsDiscrete, Compose, EnsureType, Lambda

from ..training.metric_utils import masked_accuracy, threshold_function
from ..training.optim_utils import (
    CombinedLoss,
    DeepSuperVisionLoss,
    DynUNetDirectDeepSupervision,
    IgnoreLabelDiceCELoss,
)
from ..training.surface_dice import IgnoreLabelSurfaceDiceLoss, masked_surface_dice


class SemanticSegmentationUnet(pl.LightningModule):
    """Implementation of a Unet for semantic segmentation.

    This is uses the monai Unet. See the monai docs for more details
    on the parameters.
    https://docs.monai.io/en/stable/networks.html#unet

    Parameters
    ----------
    spatial_dims : int
        The number of spatial dimensions is in the data.
    in_channels : int
        The number of channels in the input tensor.
        The default value is 1.
    out_channels : int
        The number of channels in the output tensor.
        The default value is 1.
    channels : Tuple[int, ...]
        The number of channels in each layer of the encoder/decoder.
        default=[32, 64, 128, 256, 512, 1024]
    strides : Tuple[int, ...], default=(1, 2, 2, 2, 2, 2)
        The strides for the convolutions. Must have len(channels - 1) elements.
    learning_rate : float, default=1e-2
        The learning rate for the Adam optimizer.
    min_learning_rate : float, default=1e-6
        The minimum learning rate.
    batch_size : int, default=32
        The batch size for training.
    image_key : str
        The value in the batch data dictionary containing the input image.
        Default value is "image".
    label_key : str
        The value in the batch data dictionary containing the labels.
        Default value is "label"
    roi_size : Tuple[int, ...]
        The size of the sliding window for the validation inference.
        Default value is (160, 160, 160).
    max_epochs : int, default=1000
        The maximum number of epochs for training.
    use_deep_supervision : bool, default=False
        Whether to use deep supervision.
    use_surf_dice : bool, default=False
        Whether to use surface dice loss.
    surf_dice_weight : float, default=1.0
        The weight for the surface dice loss.
    surf_dice_tokens : list, default=[]
        The tokens for which to compute the surface dice loss.

    """

    def __init__(
        self,
        spatial_dims: int = 3,
        in_channels: int = 1,
        out_channels: int = 1,
        channels: Tuple[int, ...] = [32, 64, 128, 256, 512, 1024],
        strides: Tuple[int, ...] = (1, 2, 2, 2, 2, 2),
        learning_rate: float = 1e-2,
        min_learning_rate: float = 1e-6,
        batch_size: int = 32,
        image_key: str = "image",
        label_key: str = "label",
        roi_size: Tuple[int, ...] = (160, 160, 160),
        max_epochs: int = 1000,
        use_deep_supervision: bool = False,
        use_surf_dice: bool = False,
        surf_dice_weight: float = 1.0,
        surf_dice_tokens: list = None,
    ):
        super().__init__()

        # store parameters
        self.learning_rate = learning_rate
        self.min_learning_rate = min_learning_rate
        self.batch_size = batch_size
        self.image_key = image_key
        self.label_key = label_key
        self.roi_size = roi_size
        self.max_epochs = max_epochs

        # make the network
        self._model = DynUNetDirectDeepSupervision(
            spatial_dims=spatial_dims,
            in_channels=in_channels,
            out_channels=out_channels,
            kernel_size=(3, 3, 3, 3, 3, 3),
            strides=strides,
            upsample_kernel_size=(1, 2, 2, 2, 2, 2),
            filters=channels,
            res_block=True,
            deep_supervision=True,
            deep_supr_num=2,
        )

        ### Build up loss function
        losses = []
        weights = []
        loss_inclusion_tokens = []
        ignore_dice_loss = IgnoreLabelDiceCELoss(ignore_label=2, reduction="none")
        losses.append(ignore_dice_loss)
        weights.append(1.0)
        loss_inclusion_tokens.append(["all"])  # Apply to every element

        if use_surf_dice:
            if surf_dice_tokens is None:
                surf_dice_tokens = ["all"]
            ignore_surf_dice_loss = IgnoreLabelSurfaceDiceLoss(
                ignore_label=2, soft_skel_iterations=5
            )
            losses.append(ignore_surf_dice_loss)
            weights.append(surf_dice_weight)
            loss_inclusion_tokens.append(surf_dice_tokens)

        scaled_weights = [entry / sum(weights) for entry in weights]

        loss_function = CombinedLoss(
            losses=losses,
            weights=scaled_weights,
            loss_inclusion_tokens=loss_inclusion_tokens,
        )

        self.loss_function = DeepSuperVisionLoss(
            loss_function,
            weights=(
                [1.0, 0.5, 0.25, 0.125, 0.0675]
                if use_deep_supervision
                else [1.0, 0.0, 0.0, 0.0, 0.0]
            ),
        )

        # validation metric
        self.dice_metric = DiceMetric(
            include_background=False, reduction="mean", get_not_nans=False
        )

        # transforms for val metric calculation
        threshold_function_partial = partial(threshold_function, threshold_value=0.0)
        self.post_pred = Compose(
            [
                EnsureType("tensor", device="cpu"),
                Lambda(threshold_function_partial),
                AsDiscrete(
                    to_onehot=2
                ),  # No argmax needed, since only 1 channel output
            ]
        )
        self.post_label = Compose(
            [
                EnsureType("tensor", device="cpu"),
                AsDiscrete(to_onehot=2),
            ]
        )

        self.training_step_outputs = []
        self.validation_step_outputs = []
        self.running_train_acc = 0.0
        self.running_train_surf_dice = 0.0
        self.running_val_acc = 0.0
        self.running_val_surf_dice = 0.0

    def forward(self, x) -> torch.Tensor:
        """Implementation of the forward pass.

        See the pytorch-lightning module documentation for details.
        """
        return self._model(x)

    def configure_optimizers(self):
        """Set up the Adam optimzier.

        See the pytorch-lightning module documentation for details.
        """
        # optimizer = torch.optim.Adam(self._model.parameters(), self.learning_rate)
        optimizer = torch.optim.SGD(
            self._model.parameters(),
            lr=self.learning_rate,
            momentum=0.99,
            weight_decay=3e-5,
            nesterov=True,
        )  # SGD as in nnUNet
        # scheduler = CosineAnnealingLR(
        #     optimizer, T_max=self.max_epochs, eta_min=self.min_learning_rate
        # )
        scheduler = torch.optim.lr_scheduler.LambdaLR(
            optimizer, lr_lambda=lambda epoch: (1 - epoch / self.max_epochs) ** 0.9
        )  # PolyLR from nnUNet
        return [optimizer], [scheduler]

    def training_step(
        self, batch: Dict[str, torch.Tensor], batch_idx: int
    ) -> Dict[str, float]:
        """Implementation of one training step.

        See the pytorch-lightning module documentation for details.
        """
        images, labels, ds_label = batch["image"], batch["label"], batch["dataset"]
        output = self.forward(images)
        loss = self.loss_function(output, labels, ds_label)

        stats_dict = {"train_loss": loss, "train_number": output[0].shape[0]}
        self.training_step_outputs.append(stats_dict)
        self.running_train_acc += (
            masked_accuracy(output[0], labels[0], ignore_label=2.0, threshold_value=0.0)
            * output[0].shape[0]
        )
        self.running_train_surf_dice += (
            masked_surface_dice(
                data=output[0].detach(),
                target=labels[0].detach(),
                ignore_label=2.0,
                soft_skel_iterations=5,
                smooth=1.0,
                reduction="mean",
            )
            * output[0].shape[0]
        )

        return {"loss": loss}

    def on_train_epoch_end(self):
        """What happens after each training epoch?.

        Learning rate scheduler makes one step.
        Then training loss is logged.
        """
        outputs = self.training_step_outputs
        train_loss, num_items = 0, 0
        for output in outputs:
            train_loss += output["train_loss"].sum().item() * output["train_number"]
            num_items += output["train_number"]
        mean_train_loss = torch.tensor(train_loss / num_items)

        mean_train_acc = self.running_train_acc / num_items
        mean_train_surf_dice = self.running_train_surf_dice / num_items
        self.running_train_acc = 0.0
        self.running_train_surf_dice = 0.0
        self.log("train_loss", mean_train_loss)
        self.log("train_acc", mean_train_acc)
        self.log("train_surf_dice", mean_train_surf_dice)

        self.training_step_outputs = []
        logging.info("EPOCH Training loss %s\n", mean_train_loss.item())
        logging.info("EPOCH Training acc %s\n", mean_train_acc.item())
        logging.info("EPOCH Training surface dice %s\n", mean_train_surf_dice.item())
        # Accuracy not the most informative metric, but a good sanity check
        return {"train_loss": mean_train_loss}

    def validation_step(self, batch, batch_idx):
        """Implementation of one validation step.

        This performs inference on the entire validation image
        using a sliding window. See the pytorch-lightning
        module documentation for details.
        """
        images, labels, ds_label = batch["image"], batch["label"], batch["dataset"]
        outputs = self.forward(images)
        loss = self.loss_function(outputs, labels, ds_label)

        # Cloning and adjusting preds & labels for Dice.
        # Could also use the same labels, but maybe we want to
        # compute more stats?
        outputs4dice = outputs[0].clone()
        labels4dice = labels[0].clone()
        outputs4dice[labels4dice == 2] = (
            -1.0
        )  # Setting to -1 here leads to 0-labels after thresholding
        labels4dice[labels4dice == 2] = 0  # Need to set to zero before post_label
        # Otherwise we have 3 classes
        outputs4dice = [self.post_pred(i) for i in decollate_batch(outputs4dice)]
        labels4dice = [self.post_label(i) for i in decollate_batch(labels4dice)]
        self.dice_metric(y_pred=outputs4dice, y=labels4dice)

        stats_dict = {"val_loss": loss, "val_number": outputs[0].shape[0]}
        self.validation_step_outputs.append(stats_dict)
        self.running_val_acc += (
            masked_accuracy(
                outputs[0], labels[0], ignore_label=2.0, threshold_value=0.0
            )
            * outputs[0].shape[0]
        )

        self.running_val_surf_dice += (
            masked_surface_dice(
                data=outputs[0].detach(),
                target=labels[0].detach(),
                ignore_label=2.0,
                soft_skel_iterations=5,
                smooth=1.0,
                reduction="mean",
            )
            * outputs[0].shape[0]
        )
        return stats_dict

    def on_validation_epoch_end(self):
        """Calculate validation loss/metric summary."""
        outputs = self.validation_step_outputs
        val_loss, num_items = 0, 0
        for output in outputs:
            val_loss += output["val_loss"].sum().item() * output["val_number"]
            # Need to multiply by output["val_number"] because it's later normalized.

            num_items += output["val_number"]
        mean_val_dice = self.dice_metric.aggregate().item()
        self.dice_metric.reset()
        mean_val_loss = torch.tensor(val_loss / num_items)

        mean_val_acc = self.running_val_acc / num_items
        mean_val_surf_dice = self.running_val_surf_dice / num_items
        self.running_val_acc = 0.0
        self.running_val_surf_dice = 0.0
        self.log("val_loss", mean_val_loss),
        self.log("val_dice", mean_val_dice)
        self.log("val_surf_dice", mean_val_surf_dice)
        self.log("val_accuracy", mean_val_acc)

        self.validation_step_outputs = []
        logging.info("EPOCH Validation loss %s\n", mean_val_loss.item())
        logging.info("EPOCH Validation dice %s\n", mean_val_dice)
        logging.info("EPOCH Validation surface dice %s\n", mean_val_surf_dice.item())
        logging.info("EPOCH Validation acc %s\n", mean_val_acc.item())
        return {"val_loss": mean_val_loss, "val_metric": mean_val_dice}



import os

import numpy as np

from membrain_seg.segmentation.dataloading.data_utils import (
    load_tomogram,
    make_directory_if_not_exists,
    write_nifti,
)


class InvalidCoordinatesError(Exception):
    """Exception raised of coordinates exceed tomogram boundaries."""

    pass


def pad_labels(patch, padding, pad_value=2.0):
    """
    Pads labels of a 3D array, typically for the boundaries of a 3D patch.

    Parameters
    ----------
    patch : numpy.ndarray
        The input 3D array that needs to be padded.
    padding : tuple
        The tuple containing padding dimensions for the 3D array.
        It should contain three elements: (pad_depth, pad_height, pad_width).
    pad_value: float
        Borders of the patch are padded with this value ("ignore" label)

    Returns
    -------
    patch : numpy.ndarray
        The padded 3D array.

    Notes
    -----
    This function pads the borders of the 3D array with the value of 2.0,
    typically used to ignore labels at the boundaries during a subsequent analysis.
    """
    if isinstance(padding, int):
        padding = [padding, padding, padding]
    patch[: padding[0], :, :] = pad_value
    patch[-padding[0] :, :, :] = pad_value
    patch[:, : padding[1], :] = pad_value
    patch[:, -padding[1] :, :] = pad_value
    patch[:, :, : padding[2]] = pad_value
    patch[:, :, -padding[2] :] = pad_value
    return patch


def get_out_files_and_patch_number(
    ds_token, token, out_folder_raw, out_folder_lab, patch_nr, idx_add
):
    """
    Create filenames and corrected patch numbers.

    Generates unique file names for raw and labeled data patches by incrementing patch
    number until non-existing file names are found in the specified directories.
    Also returns the final patch number used to generate these file names.

    Parameters
    ----------
    ds_token : str
        The dataset identifier used as a part of the filename.
    token : str
        The tomogram identifier used as a part of the filename.
    out_folder_raw : str
        The directory path where raw data patches are stored.
    out_folder_lab : str
        The directory path where labeled data patches are stored.
    patch_nr : int
        The initial patch number to be used for generating file names.
    idx_add : int
        The number to be added to the initial patch number to generate unique
        file names.

    Returns
    -------
    int
        The final patch number used to generate unique file names.
    str
        The full file path for the raw data patch.
    str
        The full file path for the labeled data patch.

    Notes
    -----
    The function generates filenames in the format <token>_patch<number>_raw.nii.gz
    for raw data patches and <token>_patch<number>_labels.nii.gz for labeled data
    patches.
    If a file with the same name already exists in the specified directories,
    the function increments the patch number until it finds a unique filename.
    The final patch number is obtained by adding the initial patch number and
    `idx_add` and then incrementing it further if needed.

    """
    patch_nr += idx_add
    out_file_patch = os.path.join(
        out_folder_raw, ds_token + "_" + token + "_patch" + str(patch_nr) + ".nii.gz"
    )
    out_file_patch_label = os.path.join(
        out_folder_lab, ds_token + "_" + token + "_patch" + str(patch_nr) + ".nii.gz"
    )
    exist_add = 0
    while os.path.isfile(out_file_patch):
        exist_add += 1
        out_file_patch = os.path.join(
            out_folder_raw,
            ds_token + "_" + token + "_patch" + str(patch_nr + exist_add) + ".nii.gz",
        )
        out_file_patch_label = os.path.join(
            out_folder_lab,
            ds_token + "_" + token + "_patch" + str(patch_nr + exist_add) + ".nii.gz",
        )
    return patch_nr + exist_add, out_file_patch, out_file_patch_label


def extract_patches(
    tomo_path,
    seg_path,
    coords,
    out_dir,
    ds_token="other",
    token=None,
    idx_add=0,
    pad_value=2.0,
):
    """
    Extracts 3D patches from a given tomogram and corresponding segmentation.

    The patches are then saved to the specified output directory.

    Parameters
    ----------
    tomo_path : str
        Path to the input tomogram file from which patches will be extracted.
    seg_path : str
        Path to the corresponding segmentation file.
    coords : list
        List of tuples where each tuple represents the 3D coordinates of a patch center.
    out_dir : str
        The output directory where the extracted patches will be saved.
    ds_token : str, optional
        Dataset token to uniquely identify the dataset, default is 'other'.
    token : str, optional
        Token to uniquely identify the tomogram, default is None. If None,
        the base name of the tomogram file path is used.
    idx_add : int, optional
        The index addition for patch numbering, default is 0.
    pad_value: float, optional
        Borders of extracted patch are padded with this value ("ignore" label)

    Returns
    -------
    None

    Notes
    -----
    Patches are saved in the 'imagesCorr' and 'labelsCorr' subdirectories of the
    output directory.
    The patch from the tomogram and the corresponding patch from the segmentation
    are saved with the same name for easy correspondence.

    Exceptions
    ----------
    Raises InvalidCoordinatesError if a patch cannot be extracted due to the provided
    coordinates being too close to the border of the tomogram.
    """
    padding = (16, 16, 16)
    if token is None:
        token = os.path.splitext(os.path.basename(tomo_path))[0]
    out_folder_raw = os.path.join(out_dir, "imagesCorr")
    out_folder_lab = os.path.join(out_dir, "labelsCorr")
    make_directory_if_not_exists(out_folder_raw)
    make_directory_if_not_exists(out_folder_lab)

    tomo = load_tomogram(tomo_path).data
    labels = load_tomogram(seg_path).data

    for patch_nr, cur_coords in enumerate(coords):
        patch_nr, out_file_patch, out_file_patch_label = get_out_files_and_patch_number(
            ds_token, token, out_folder_raw, out_folder_lab, patch_nr, idx_add
        )
        print("Extracting patch nr", patch_nr, "from tomo", token)
        try:
            min_coords = np.array(cur_coords) - 80
            if not (
                np.all(min_coords) >= 0
                and np.all(min_coords < (np.array(tomo.shape) - 160))
            ):
                raise InvalidCoordinatesError(
                    "Invalid coordinates: Some values are outside the allowed range.\
                Your selected center points should be at least 80 voxels away from \
                the tomogram borders."
                )
            cur_patch = tomo[
                min_coords[0] : min_coords[0] + 160,
                min_coords[1] : min_coords[1] + 160,
                min_coords[2] : min_coords[2] + 160,
            ]
            cur_patch_labels = labels[
                min_coords[0] : min_coords[0] + 160,
                min_coords[1] : min_coords[1] + 160,
                min_coords[2] : min_coords[2] + 160,
            ].copy()
            cur_patch_labels = pad_labels(
                cur_patch_labels, padding, pad_value=pad_value
            )

            cur_patch = np.transpose(cur_patch, (2, 1, 0))
            cur_patch_labels = np.transpose(cur_patch_labels, (2, 1, 0))
            write_nifti(out_file_patch, cur_patch)
            write_nifti(out_file_patch_label, cur_patch_labels)

        except InvalidCoordinatesError as e:
            print("Error:", str(e))



import logging
from typing import Optional

from typer import Option

from .cli import OPTION_PROMPT_KWARGS as PKWARGS
from .cli import cli

logging.basicConfig(level=logging.INFO)


@cli.command(name="download_data", no_args_is_help=True)
def download_data(
    out_folder: str = Option(  # noqa: B008
        ...,
        help="Folder to store the downloaded MemBrain-seg dataset (14GB .zip file).",
        **PKWARGS,
    ),
    num_parallel_processes: int = Option(  # noqa: B008
        help='Number of parallel processes to use for downloading the dataset.\
            Since dataset is hosted on Zenodo with limited bandwidth,\
            using multiple parallel processes can speed up the download.\
            Pass an integer value greater than 0. For example, to use 4 parallel\
            processes, pass "4". Default is "1" (single process).',
        default=1,
    ),
):
    """
    CLI for downloading the MemBrain-seg dataset.

    Downloads the MemBrain-seg benchmark dataset from Zenodo
    and saves it to the specified output folder.

    The dataset can be used to re-train MemBrain-seg, but also to
    benchmark MemBrain-seg against other segmentation methods.

    Parameters
    ----------
    out_folder : str
        Folder to store the downloaded MemBrain-seg dataset (14GB .zip file).
    num_parallel_processes : str
        Number of parallel processes to use for downloading the dataset.
        Since dataset is hosted on Zenodo with limited bandwidth,
        using multiple parallel processes can speed up the download.
        Pass an integer value greater than 0. For example, to use 4 parallel
        processes, pass "4". Default is "1" (single process).
    """
    from membrain_seg.benchmark.download_dataset import download_data as _download_data

    _download_data(out_folder=out_folder, num_parallel_processes=num_parallel_processes)


@cli.command(name="benchmark", no_args_is_help=True)
def benchmark(
    pred_path: str = Option(  # noqa: B008
        ...,
        help="Folder containing the predicted segmentation results to benchmark.\
            Should contain .mrc files with predicted segmentations per membrane patch.\
            See documentation for details.",
        **PKWARGS,
    ),
    gt_path: str = Option(  # noqa: B008
        ...,
        help="Folder containing the ground truth segmentations.\
            If not done yet, download the MemBrain-seg dataset using the\
            'download_data' command. ",
        **PKWARGS,
    ),
    out_dir: str = Option(  # noqa: B008
        ...,
        help="Output directory to store the benchmarking results (CSV file and plots).",
        **PKWARGS,
    ),
    out_file_token: Optional[str] = Option(  # noqa: B008
        "stats",
        help="Optional token to append to the output filenames.",
    ),
    skeletonization_method: Optional[str] = Option(  # noqa: B008
        "3D-NMS",
        help='Skeletonization method to use. Supported: "3D-NMS", "2D-skimage".',
    ),
):
    """
    CLI for benchmarking with MemBrain-seg's annotated dataset.

    Compares predicted segmentations against ground truth segmentations
    using Dice score and Surface-Dice score metrics.
    Saves the benchmarking results as a CSV file.


    Parameters
    ----------
    pred_path : str
        Folder containing the predicted segmentation results to benchmark.
        Should contain .mrc files with predicted segmentations per membrane patch.
        See documentation for details.
    gt_path : str
        Folder containing the ground truth segmentations.
        If not done yet, download the MemBrain-seg dataset using the
        'download_data' command.
    out_dir : str
        Output directory to store the benchmarking results (CSV file).
    out_file_token : Optional[str]
        Optional token to append to the output filenames.
    skeletonization_method : str
        Skeletonization method to use. Supported: "3D-NMS", "2D-skimage".

    Examples
    --------
    membrain benchmark --pred-path <path-to-predictions>
        --gt-path <path-to-ground-truth> --out-dir <output-directory>
        --out-file-token <optional-token>

    """
    from membrain_seg.benchmark.compute_stats import (
        compute_stats,
    )

    compute_stats(
        dir_pred=pred_path,
        dir_gt=gt_path,
        out_dir=out_dir,
        out_file_token=out_file_token,
        skeletonization_method=skeletonization_method,
    )



"""
Surface Dice implementation.

Adapted from: clDice - A Novel Topology-Preserving Loss Function for Tubular 
Structure Segmentation
Original Authors: Johannes C. Paetzold and Suprosanna Shit
Sources: https://github.com/jocpae/clDice/blob/master/cldice_loss/pytorch/
        soft_skeleton.py
         https://github.com/jocpae/clDice/blob/master/cldice_loss/pytorch/cldice.py 
License: MIT License.

The following code is a modification of the original clDice implementation.
Modifications were made to include additional functionality and integrate 
with new project requirements. The original license and copyright notice are 
provided below.

MIT License

Copyright (c) 2021 Johannes C. Paetzold and Suprosanna Shit

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
"""

import math

import torch
import torch.nn.functional as F
from torch.nn.functional import sigmoid
from torch.nn.modules.loss import _Loss


def soft_erode(img: torch.Tensor, separate_pool: bool = False) -> torch.Tensor:
    """
    Apply soft erosion operation to the input image.

    Soft erosion is achieved by applying a min-pooling operation to the input image.

    Parameters
    ----------
    img : torch.Tensor
        Input image tensor with shape (B, C, D, H, W)
    separate_pool : bool, optional
        If True, perform separate 3D max-pooling operations along different axes.
        Default is False.

    Returns
    -------
    torch.Tensor
        Eroded image tensor with the same shape as the input.

    Raises
    ------
    ValueError
        If the input tensor has an unsupported number of dimensions.

    Notes
    -----
    - The soft erosion can be performed with separate 3D min-pooling operations
            along different axes if separate_pool is True, or with a single
            3D min-pooling operation with a kernel of size (3, 3, 3) if
            separate_pool is False.
    """
    assert len(img.shape) == 5
    if separate_pool:
        p1 = -F.max_pool3d(-img, (3, 1, 1), (1, 1, 1), (1, 0, 0))
        p2 = -F.max_pool3d(-img, (1, 3, 1), (1, 1, 1), (0, 1, 0))
        p3 = -F.max_pool3d(-img, (1, 1, 3), (1, 1, 1), (0, 0, 1))
        return torch.min(torch.min(p1, p2), p3)
    p4 = -F.max_pool3d(-img, (3, 3, 3), (1, 1, 1), (1, 1, 1))
    return p4


def soft_dilate(img: torch.Tensor) -> torch.Tensor:
    """
    Apply soft dilation operation to the input image.

    Soft dilation is achieved by applying a max-pooling operation to the input image.

    Parameters
    ----------
    img : torch.Tensor
        Input image tensor with shape (B, C, D, H, W).

    Returns
    -------
    torch.Tensor
        Dilated image tensor with the same shape as the input.

    Raises
    ------
    ValueError
        If the input tensor has an unsupported number of dimensions.

    Notes
    -----
    - For 5D input, the soft dilation is performed using a 3D max-pooling operation
            with a kernel of size (3, 3, 3).
    """
    assert len(img.shape) == 5
    return F.max_pool3d(img, (3, 3, 3), (1, 1, 1), (1, 1, 1))


def soft_open(img: torch.Tensor, separate_pool: bool = False) -> torch.Tensor:
    """
    Apply soft opening operation to the input image.

    Soft opening is achieved by applying soft erosion followed by soft dilation.
    The intention of soft opening is to remove thin membranes from the segmentation.

    Parameters
    ----------
    img : torch.Tensor
        Input image tensor with shape (B, C, D, H, W).
    separate_pool : bool, optional
        If True, perform separate erosion and dilation operations. Default is False.

    Returns
    -------
    torch.Tensor
        Opened image tensor with the same shape as the input.

    Notes
    -----
    - Soft opening is performed by applying soft erosion followed by soft dilation
            to the input image.
    - For 5D input, separate erosion and dilation can be performed if separate_pool
            is True.
    """
    return soft_dilate(soft_erode(img, separate_pool=separate_pool))


def soft_skel(
    img: torch.Tensor, iter_: int, separate_pool: bool = False
) -> torch.Tensor:
    """
    Compute the soft skeleton of the input image.

    The skeleton is computed by applying soft erosion iteratively to the input image.
    In each iteration, the difference between the input image and the "opened" image is
    computed and added to the skeleton.

    Reasoning: if there is a difference between the input image and the "opened" image,
    there must be a thin membrane skeleton in the input image that was removed by the
    opening operation.

    Parameters
    ----------
    img : torch.Tensor
        Input image tensor with shape (B, C, D, H, W).
    iter_ : int
        Number of iterations for skeletonization.
    separate_pool : bool, optional
        If True, perform separate erosion and dilation operations.
        Default is False.

    Returns
    -------
    torch.Tensor
        Soft skeleton image tensor with the same shape as the input.

    Notes
    -----
    - Separate erosion can be performed if separate_pool is True.
    """
    img1 = soft_open(img, separate_pool=separate_pool)
    skel = F.relu(img - img1)
    for _j in range(iter_):
        img = soft_erode(img)
        img1 = soft_open(img, separate_pool=separate_pool)
        delta = F.relu(img - img1)
        skel = skel + F.relu(delta - skel * delta)
    return skel


def gaussian_kernel(size: int, sigma: float) -> torch.Tensor:
    """
    Creates a 3D Gaussian kernel using the specified size and sigma.

    Parameters
    ----------
    size : int
        The size of the Gaussian kernel. It determines the length of
        each dimension of the cube.
    sigma : float
        The standard deviation of the Gaussian kernel. It controls
        the spread of the Gaussian.

    Returns
    -------
    torch.Tensor
        A 3D tensor representing the Gaussian kernel.

    Notes
    -----
    The function creates a Gaussian kernel, which is essentially a
    cube of dimensions [size, size, size]. Each entry in the cube is
    computed using the Gaussian function based on its distance from the center.
    The kernel is normalized so that its total sum equals 1.
    """
    # Define a coordinate grid centered at (0,0,0)
    grid = torch.arange(size, dtype=torch.float32) - (size - 1) / 2
    # Create a 3D meshgrid
    x, y, z = torch.meshgrid(grid, grid, grid)
    xyz_grid = torch.stack([x, y, z], dim=-1)

    # Calculate the 3D Gaussian kernel
    gaussian_kernel = torch.exp(-torch.sum(xyz_grid**2, dim=-1) / (2 * sigma**2))
    gaussian_kernel /= (2 * math.pi * sigma**2) ** (3 / 2)  # Normalize

    # Ensure sum of values in gaussian kernel equals 1.
    gaussian_kernel = gaussian_kernel / torch.sum(gaussian_kernel)
    return gaussian_kernel


gaussian_kernel_dict = {}
""" Not sure why, but moving the gaussian kernel to GPU takes surprisingly long.+
So we precompute it, store it on GPU, and reuse it.
"""


def apply_gaussian_filter(
    seg: torch.Tensor, kernel_size: int, sigma: float
) -> torch.Tensor:
    """
    Apply a Gaussian filter to a segmentation tensor using PyTorch.

    This function convolves the input tensor with a Gaussian kernel.
    The function creates or retrieves a Gaussian kernel based on the
    specified size and standard deviation, and applies 3D convolution to each
    channel of each batch item with appropriate padding to maintain spatial
    dimensions.

    Parameters
    ----------
    seg : torch.Tensor
        The input segmentation tensor of shape (batch, channel, X, Y, Z).
    kernel_size : int
        The size of the Gaussian kernel, determining the length of each
        dimension of the cube.
    sigma : float
        The standard deviation of the Gaussian kernel, controlling the spread.

    Returns
    -------
    torch.Tensor
        The filtered segmentation tensor of the same shape as input.

    Notes
    -----
    This function uses a precomputed dictionary to enhance performance by
    storing Gaussian kernels. If a kernel with the specified size and standard
    deviation does not exist in the dictionary, it is created and added. The
    function assumes the input tensor is a 5D tensor, applies 3D convolution
    using the Gaussian kernel with padding to maintain spatial dimensions, and
    it performs the operation separately for each channel of each batch item.
    """
    # Create the Gaussian kernel or load it from the dictionary
    if (kernel_size, sigma) not in gaussian_kernel_dict.keys():
        gaussian_kernel_dict[(kernel_size, sigma)] = gaussian_kernel(
            kernel_size, sigma
        ).to(seg.device)
    g_kernel = gaussian_kernel_dict[(kernel_size, sigma)]

    # Add batch and channel dimensions
    g_kernel = g_kernel.view(1, 1, *g_kernel.size())
    # Apply the Gaussian filter to each channel
    padding = kernel_size // 2

    # Move the kernel to the same device as the segmentation tensor
    g_kernel = g_kernel.to(seg.device)

    # Apply the Gaussian filter
    filtered_seg = F.conv3d(seg, g_kernel, padding=padding, groups=seg.shape[1])
    return filtered_seg


def get_GT_skeleton(gt_seg: torch.Tensor, iterations: int = 5) -> torch.Tensor:
    """
    Generate the skeleton of a ground truth segmentation.

    This function takes a ground truth segmentation `gt_seg`, smooths it using a
    Gaussian filter, and then computes its soft skeleton using the `soft_skel` function.

    Intention: When using the binary ground truth segmentation for skeletonization,
    the resulting skeleton is very patchy and not smooth. When using the smoothed
    ground truth segmentation, the resulting skeleton is much smoother and more
    accurate.

    Parameters
    ----------
    gt_seg : torch.Tensor
        A torch.Tensor representing the ground truth segmentation.
        Shape: (B, C, D, H, W)
    iterations : int, optional
        The number of iterations for skeletonization. Default is 5.

    Returns
    -------
    torch.Tensor
        A torch.Tensor representing the skeleton of the ground truth segmentation.

    Notes
    -----
    - The input `gt_seg` should be a binary segmentation tensor where 1 represents the
        object of interest.
    - The function first smooths the `gt_seg` using a Gaussian filter to enhance the
        object's structure.
    - The skeletonization process is performed using the `soft_skel` function with the
        specified number of iterations.
    - The resulting skeleton is returned as a binary torch.Tensor where 1 indicates the
        skeleton points.
    """
    gt_smooth = (
        apply_gaussian_filter((gt_seg == 1) * 1.0, kernel_size=15, sigma=2.0) * 1.5
    )
    skel_gt = soft_skel(gt_smooth, iter_=iterations)
    return skel_gt


def masked_surface_dice(
    data: torch.Tensor,
    target: torch.Tensor,
    ignore_label: int = 2,
    soft_skel_iterations: int = 3,
    smooth: float = 3.0,
    binary_prediction: bool = False,
    reduction: str = "none",
) -> torch.Tensor:
    """
    Compute the surface Dice loss with masking for ignore labels.

    The surface Dice loss measures the similarity between the predicted segmentation's
    skeleton and the ground truth segmentation (and vice versa). Labels annotated with
    "ignore_label" are ignored.

    Parameters
    ----------
    data : torch.Tensor
        Tensor of model outputs representing the predicted segmentation.
        Expected shape: (B, C, D, H, W)
    target : torch.Tensor
        Tensor of target labels representing the ground truth segmentation.
        Expected shape: (B, 1, D, H, W)
    ignore_label : int
        The label value to be ignored when computing the loss.
    soft_skel_iterations : int
        Number of iterations for skeletonization in the underlying operations.
    smooth : float
        Smoothing factor to avoid division by zero.
    binary_prediction : bool
        If True, the predicted segmentation is assumed to be binary. Default is False.
    reduction : str
        Specifies the reduction to apply to the output. Default is "none".

    Returns
    -------
    torch.Tensor
        The calculated surface Dice loss.
    """
    # Create a mask to ignore the specified label in the target
    data = sigmoid(data)
    mask = target != ignore_label

    # Compute soft skeletonization
    if binary_prediction:
        skel_pred = get_GT_skeleton(data.clone(), soft_skel_iterations)
    else:
        skel_pred = soft_skel(data.clone(), soft_skel_iterations, separate_pool=False)
    skel_true = get_GT_skeleton(target.clone(), soft_skel_iterations)

    # Mask out ignore labels
    skel_pred[~mask] = 0
    skel_true[~mask] = 0

    # compute surface dice loss
    tprec = (
        torch.sum(torch.multiply(skel_pred, target), dim=(1, 2, 3, 4)) + smooth
    ) / (torch.sum(skel_pred, dim=(1, 2, 3, 4)) + smooth)
    tsens = (torch.sum(torch.multiply(skel_true, data), dim=(1, 2, 3, 4)) + smooth) / (
        torch.sum(skel_true, dim=(1, 2, 3, 4)) + smooth
    )
    surf_dice_loss = 2.0 * (tprec * tsens) / (tprec + tsens)
    if reduction == "none":
        return surf_dice_loss
    elif reduction == "mean":
        return torch.mean(surf_dice_loss)


class IgnoreLabelSurfaceDiceLoss(_Loss):
    """
    Surface Dice loss, adding ignore labels.

    Parameters
    ----------
    ignore_label : int
        The label to ignore when calculating the loss.
    reduction : str, optional
        Specifies the reduction to apply to the output, by default "mean".
    kwargs : dict
        Additional keyword arguments.
    """

    def __init__(
        self,
        ignore_label: int,
        soft_skel_iterations: int = 3,
        smooth: float = 3.0,
        **kwargs,
    ) -> None:
        super().__init__()
        self.ignore_label = ignore_label
        self.soft_skel_iterations = soft_skel_iterations
        self.smooth = smooth

    def forward(self, data: torch.Tensor, target: torch.Tensor) -> torch.Tensor:
        """
        Compute the loss.

        Parameters
        ----------
        data : torch.Tensor
            Tensor of model outputs.
            Expected shape: (B, C, D, H, W)
        target : torch.Tensor
            Tensor of target labels.
            Expected shape: (B, 1, D, H, W)

        Returns
        -------
        torch.Tensor
            The calculated loss.
        """
        # Create a mask to ignore the specified label in the target
        surf_dice_score = masked_surface_dice(
            data=data,
            target=target,
            ignore_label=self.ignore_label,
            soft_skel_iterations=self.soft_skel_iterations,
            smooth=self.smooth,
        )
        surf_dice_loss = 1.0 - surf_dice_score
        return surf_dice_loss



from typing import Optional

import torch


def masked_accuracy(
    y_pred: torch.Tensor,
    y_gt: torch.Tensor,
    ignore_label: Optional[int] = None,
    threshold_value: float = 0.0,
) -> torch.Tensor:
    """
    Computes fraction of correctly predicted voxels after thresholding.

    If an ignore label is provided, accuracy is only computed for voxels
    where the GT label is NOT equal to the ignore label.

    Parameters
    ----------
    y_pred : torch.Tensor
        Tensor representing the model's predictions.
    y_gt : torch.Tensor
        Tensor representing the ground truth labels.
    ignore_label : Optional[int], optional
        The label to ignore when calculating accuracy, by default None
    threshold_value : float, optional
        The threshold value to convert predictions into binary values, by default 0.0

    Returns
    -------
    torch.Tensor
        The computed accuracy of the model's predictions.
    """
    mask = (
        y_gt == ignore_label
        if ignore_label is not None
        else torch.zeros_like(y_gt).bool()
    )
    acc = (threshold_function(y_pred, threshold_value=threshold_value) == y_gt).float()
    acc[mask] = 0.0
    acc = acc.sum()
    acc /= (~mask).sum()
    return acc


def threshold_function(
    predictions: torch.Tensor, threshold_value: float = 0.0
) -> torch.Tensor:
    """
    Return 0-1 Array from thresholding.

    Parameters
    ----------
    predictions : torch.Tensor
        The tensor containing prediction values.
    threshold_value : float, optional
        The threshold value for converting predictions into binary values,
        by default 0.0

    Returns
    -------
    torch.Tensor
        The binary tensor obtained after thresholding the predictions.
    """
    binary_mask = (predictions >= threshold_value).float()
    return binary_mask
